"""String utilities for the userstring type."""

import re

from .exceptions import IterableTooLong, _raise_in_context

__all__ = ("check_format_spec", "FORMAT_SPEC_RE", "PRINTF_TEMPLATE_RE", "JoinProxy", "TranslateTableProxy")

# ==== format spec ====
# .format()-style
# https://docs.python.org/3/library/string.html#format-specification-mini-language
_FILL = r"."
_ALIGN = r"[<>=^]"
_SIGN = r"[+\- ]"
_WIDTH = r"\d+"
_GROUPING_OPTION = r"[_,]"
_PRECISION = r"\d+"
_TYPE = r"[bcdeEfFgGnosxX%]"
FORMAT_SPEC_RE = re.compile(
    rf"(?:(?P<fill>{_FILL})?(?P<align>{_ALIGN}))?"
    rf"(?P<sign>{_SIGN})?"
    rf"(?P<alt_form>#)?"
    rf"(?P<zero_pad>0)?"
    rf"(?P<width>{_WIDTH})?"
    rf"(?P<grouping_option>{_GROUPING_OPTION})?"
    rf"(?:\.(?P<precision>{_PRECISION}))?"
    rf"(?P<type>{_TYPE})?"
)


def check_format_spec(config, format_spec):
    # validate that the format string is safe
    match = FORMAT_SPEC_RE.match(format_spec)
    if not match:
        raise ValueError("Invalid format specifier")

    precision_len = 0
    w = match.group("width")
    p = match.group("precision")
    if w:
        precision_len += int(w)
    if p:
        precision_len += int(p)
    if precision_len > config.max_const_len:
        _raise_in_context(IterableTooLong, "This str is too large")


# printf-style
# https://docs.python.org/3/library/stdtypes.html#printf-style-string-formatting
_PF_MAPPING_KEY = r"[^)]*"  # any sequence of non-")"
_PF_CONVERSION_FLAGS = r"[#0\- +]"
_PF_WIDTH = r"\*|\d+"
_PF_PRECISION = r"\*|\d+"
_PF_LENGTH_MODIFIER = r"[hlL]"
_PF_TYPE = r"[diouxXeEfFgGcrsa%]"
PRINTF_TEMPLATE_RE = re.compile(
    rf"%"
    rf"(?:\((?P<mapping_key>{_PF_MAPPING_KEY})\))?"
    rf"(?P<conversion_flags>{_PF_CONVERSION_FLAGS})*"
    rf"(?P<width>{_PF_WIDTH})?"
    rf"(?:\.(?P<precision>{_PF_PRECISION}))?"
    rf"(?P<length_modifier>{_PF_LENGTH_MODIFIER})?"
    rf"(?P<type>{_PF_TYPE})"
)


# ==== helpers ====
class JoinProxy:
    """
    A helper to return the right types for str.join().
    If the sequence would return a userstring, returns it as a base str instead.
    """

    def __init__(self, str_type, seq):
        self._str_type = str_type
        self.seq = seq

    def __iter__(self):
        for item in self.seq:
            if isinstance(item, self._str_type):
                yield str(item)
            else:
                yield item


class TranslateTableProxy:
    """
    A helper to return the right types for str.translate().
    If the table would return a userstring, returns it as a base str instead.
    """

    def __init__(self, str_type, table):
        self._str_type = str_type
        self.table = table

    def __getitem__(self, item):
        out = self.table[item]
        if isinstance(out, self._str_type):  # translate() expects strictly a str type
            return str(out)
        return out



from avrae_ls.symbols import build_symbol_table, document_symbols, find_definition_range


def test_build_symbol_table_and_document_symbols():
    source = "x = 1\n\ndef foo():\n    return x\n"
    table = build_symbol_table(source)

    assert any(entry.name == "x" for entry in table.entries)
    assert any(entry.name == "foo" for entry in table.entries)

    doc_symbols = document_symbols(source)
    assert {s.name for s in doc_symbols} == {"x", "foo"}


def test_find_definition_range():
    source = "x = 1\ny = x + 2\n"
    table = build_symbol_table(source)
    rng = find_definition_range(table, "x")
    assert rng is not None
    assert rng.start.line == 0
    assert rng.start.character == 0



import ast
import operator as op
from typing import Sequence

from .exceptions import *
from .types import *

__all__ = ("DraconicConfig", "OperatorMixin", "zip_star")

# ===== config =====
DISALLOW_PREFIXES = ["_", "func_"]
DISALLOW_METHODS = ["format", "format_map", "mro", "tb_frame", "gi_frame", "ag_frame", "cr_frame", "exec"]


class DraconicConfig:
    """A configuration object to pass into the Draconic interpreter."""

    def __init__(
        self,
        max_const_len=200_000,
        max_loops=10_000,
        max_statements=100_000,
        max_power_base=1_000_000,
        max_power=1_000,
        disallow_prefixes=None,
        disallow_methods=None,
        default_names=None,
        builtins_extend_default=True,
        max_int_size=64,
        max_recursion_depth=50,
    ):
        """
        Configuration object for the Draconic interpreter.

        :param int max_const_len: The maximum length literal that should be allowed to be constructed.
        :param int max_loops: The maximum total number of loops allowed per execution.
        :param int max_statements: The maximum total number of statements allowed per execution.
        :param int max_power_base: The maximum power base (x in x ** y)
        :param int max_power: The maximum power (y in x ** y)
        :param list disallow_prefixes: A list of str - attributes starting with any of these will be inaccessible
        :param list disallow_methods: A list of str - methods named these will not be callable
        :param dict default_names: A dict of str: Any - default names in the runtime
        :param bool builtins_extend_default: If False, ``builtins`` to the interpreter overrides default names
        :param int max_int_size: The maximum allowed size of integers (-2^(pow-1) to 2^(pow-1)-1). Default 64.
                                 Integers can technically reach up to double this size before size check.
                                 *Not* the max value!
        :param int max_recursion_depth: The maximum allowed recursion depth.
        """
        if disallow_prefixes is None:
            disallow_prefixes = DISALLOW_PREFIXES
        if disallow_methods is None:
            disallow_methods = DISALLOW_METHODS

        self.max_const_len = max_const_len
        self.max_loops = max_loops
        self.max_statements = max_statements
        self.max_power_base = max_power_base
        self.max_power = max_power
        self.max_int_size = max_int_size
        self.min_int = -(2 ** (max_int_size - 1))
        self.max_int = (2 ** (max_int_size - 1)) - 1
        self.disallow_prefixes = disallow_prefixes
        self.disallow_methods = disallow_methods
        self.builtins_extend_default = builtins_extend_default
        self.max_recursion_depth = max_recursion_depth

        # types
        self._list = safe_list(self)
        self._dict = safe_dict(self)
        self._set = safe_set(self)
        self._str = safe_str(self)

        # default names
        if default_names is None:
            default_names = self._default_names()
        self.default_names = default_names

    @property
    def list(self):
        return self._list

    @property
    def dict(self):
        return self._dict

    @property
    def set(self):
        return self._set

    @property
    def str(self):
        return self._str

    def _default_names(self):
        return {
            "True": True,
            "False": False,
            "None": None,
            # functions
            "bool": bool,
            "int": int,
            "float": float,
            "str": self.str,
            "tuple": tuple,
            "dict": self.dict,
            "list": self.list,
            "set": self.set,
        }


# ===== operators =====
class OperatorMixin:
    """A mixin class to provide the operators."""

    def __init__(self, config):
        """
        :type config: draconic.helpers.DraconicConfig
        """
        self._config = config

        self.operators = {
            # binary
            ast.Add: self._safe_add,
            ast.Sub: self._safe_sub,
            ast.Mult: self._safe_mult,
            ast.Div: op.truediv,
            ast.FloorDiv: op.floordiv,
            ast.Pow: self._safe_power,
            ast.Mod: op.mod,
            ast.LShift: self._safe_lshift,
            ast.RShift: op.rshift,
            ast.BitOr: op.or_,
            ast.BitXor: op.xor,
            ast.BitAnd: op.and_,
            ast.Invert: op.invert,
            # unary
            ast.Not: op.not_,
            ast.USub: op.neg,
            ast.UAdd: op.pos,
            # comparison
            ast.Eq: op.eq,
            ast.NotEq: op.ne,
            ast.Gt: op.gt,
            ast.Lt: op.lt,
            ast.GtE: op.ge,
            ast.LtE: op.le,
            ast.In: lambda x, y: op.contains(y, x),
            ast.NotIn: lambda x, y: not op.contains(y, x),
            ast.Is: lambda x, y: x is y,
            ast.IsNot: lambda x, y: x is not y,
        }

    def _safe_power(self, a, b):
        """Exponent: limit power base and power to prevent CPU-locking computation"""
        if abs(a) > self._config.max_power_base or abs(b) > self._config.max_power:
            _raise_in_context(NumberTooHigh, f"{a} ** {b} is too large of an exponent")
        result = a**b
        if isinstance(result, int) and (result < self._config.min_int or result > self._config.max_int):
            _raise_in_context(NumberTooHigh, "This exponent would create a number too large")
        return result

    def _safe_mult(self, a, b):
        """Multiplication: limit the size of iterables that can be created, and the max size of ints"""
        # sequences can only be multiplied by ints, so this is safe
        self._check_binop_operands(a, b)
        if isinstance(b, int) and b * approx_len_of(a) > self._config.max_const_len:
            _raise_in_context(IterableTooLong, "Multiplying these two would create something too long")
        if isinstance(a, int) and a * approx_len_of(b) > self._config.max_const_len:
            _raise_in_context(IterableTooLong, "Multiplying these two would create something too long")
        result = a * b
        if isinstance(result, int) and (result < self._config.min_int or result > self._config.max_int):
            _raise_in_context(NumberTooHigh, "Multiplying these two would create a number too large")
        return result

    def _safe_add(self, a, b):
        """Addition: limit the size of iterables that can be created, and the max size of ints"""
        self._check_binop_operands(a, b)
        if approx_len_of(a) + approx_len_of(b) > self._config.max_const_len:
            _raise_in_context(IterableTooLong, "Adding these two would create something too long")
        result = a + b
        if isinstance(result, int) and (result < self._config.min_int or result > self._config.max_int):
            _raise_in_context(NumberTooHigh, "Adding these two would create a number too large")
        return result

    def _safe_sub(self, a, b):
        """Addition: limit the max size of ints"""
        self._check_binop_operands(a, b)
        result = a - b
        if isinstance(result, int) and (result < self._config.min_int or result > self._config.max_int):
            _raise_in_context(NumberTooHigh, "Subtracting these two would create a number too large")
        return result

    def _safe_lshift(self, a, b):
        """Left Bit-Shift: limit the size of integers/floats to prevent CPU-locking computation"""
        self._check_binop_operands(a, b)

        if isinstance(b, int) and b > self._config.max_int_size - 2:
            _raise_in_context(NumberTooHigh, f"{a} << {b} is too large of a shift")

        result = a << b
        if isinstance(result, int) and (result < self._config.min_int or result > self._config.max_int):
            _raise_in_context(NumberTooHigh, "Shifting these two would create a number too large")

        return a << b

    def _check_binop_operands(self, a, b):
        """Ensures both operands of a binary operation are safe (int limit)."""
        if isinstance(a, int) and (a < self._config.min_int or a > self._config.max_int):
            _raise_in_context(NumberTooHigh, "This number is too large")
        if isinstance(b, int) and (b < self._config.min_int or b > self._config.max_int):
            _raise_in_context(NumberTooHigh, "This number is too large")


# ==== other utils ====
def zip_star(a: Sequence, b: Sequence, star_index: int):
    """
    Like zip(a, b), but zips the element at ``a[star_index]`` with a list of 0..len(b) elements such that every other
    element of ``a`` maps to exactly one element of ``b``.

    >>> zip_star(['a', 'b', 'c'], [1, 2, 3, 4], star_index=1)  # like a, *b, c = [1, 2, 3, 4]
    [('a', 1), ('b', [2, 3]), ('c', 4)]
    >>> zip_star(['a', 'b', 'c'], [1, 2], star_index=1)  # like a, *b, c = [1, 2]
    [('a', 1), ('b', []), ('c', 2)]

    Requires ``len(b) >= len(a) - 1`` and ``star_index < len(a)``.
    """
    if not 0 <= star_index < len(a):
        raise IndexError("'star_index' must be a valid index of 'a'")
    if not len(b) >= len(a) - 1:
        raise ValueError("'b' must be no more than 1 shorter than 'a'")

    length_difference = len(b) - (len(a) - 1)

    yield from zip(a[:star_index], b[:star_index])
    yield a[star_index], b[star_index : star_index + length_difference]
    yield from zip(a[star_index + 1 :], b[star_index + length_difference :])



import pytest
from pathlib import Path

from avrae_ls.config import AvraeLSConfig
from avrae_ls.context import ContextBuilder, GVarResolver
from avrae_ls.server import AvraeLanguageServer, refresh_gvars


@pytest.mark.asyncio
async def test_refresh_resets_and_seeds_without_fetch():
    cfg = AvraeLSConfig.default(Path("."))
    resolver = GVarResolver(cfg)
    resolver.reset({"old": "value"})

    snapshot = await resolver.refresh({"seed": "x"}, keys=["missing"])

    assert snapshot.get("seed") == "x"
    assert "old" not in snapshot
    # enable_gvar_fetch defaults to false, so missing key should not be added
    assert "missing" not in snapshot


@pytest.mark.asyncio
async def test_refresh_gvars_command_uses_profile_seed():
    server = AvraeLanguageServer()
    server.load_workspace(Path("."))
    result = await refresh_gvars(server, {"profile": "default", "keys": ["abc123"]})

    assert "gvars" in result
    assert result["count"] == len(result["gvars"])
    assert "abc123" not in result["gvars"]


def test_context_builder_preserves_cached_gvars(tmp_path):
    cfg = AvraeLSConfig.default(tmp_path)
    builder = ContextBuilder(cfg)
    resolver = builder.gvar_resolver

    resolver.seed({"cached": "value"})
    builder.build()
    builder.build("default")

    assert resolver.get_local("cached") == "value"



from __future__ import annotations

import argparse

from .server import create_server


def main(argv: list[str] | None = None) -> None:
    parser = argparse.ArgumentParser(description="Avrae draconic alias language server")
    parser.add_argument("--tcp", action="store_true", help="Run in TCP mode instead of stdio")
    parser.add_argument("--host", default="127.0.0.1", help="TCP host (when --tcp is set)")
    parser.add_argument("--port", type=int, default=2087, help="TCP port (when --tcp is set)")
    parser.add_argument("--stdio", action="store_true", help="Accept stdio flag for VS Code clients (ignored)")
    args = parser.parse_args(argv)

    server = create_server()
    if args.tcp:
        server.start_tcp(args.host, args.port)
    else:
        server.start_io()


if __name__ == "__main__":
    main()



import collections
import itertools
import re
import string
from typing import Iterator


class BadArgument(Exception):
    pass


class ExpectedClosingQuoteError(Exception):
    def __init__(self, quote):
        super().__init__(quote)
        self.quote = quote

    def __str__(self):
        return f"Expected closing quote {self.quote}"


class InvalidArgument(Exception):
    pass


class StringView:
    """Minimal stand-in for disnake.ext.commands.view.StringView used by argparser."""

    def __init__(self, buffer: str):
        self.buffer = buffer
        self.index = 0

    @property
    def current(self):
        if self.eof:
            return None
        return self.buffer[self.index]

    @property
    def eof(self) -> bool:
        return self.index >= len(self.buffer)

    def skip_ws(self):
        while not self.eof and self.buffer[self.index].isspace():
            self.index += 1

    def get(self):
        if self.eof:
            return None
        ch = self.buffer[self.index]
        self.index += 1
        return ch

    def undo(self):
        if self.index > 0:
            self.index -= 1


EPHEMERAL_ARG_RE = re.compile(r"(\S+)(\d+)")
SINGLE_ARG_RE = re.compile(r"([a-zA-Z]\S*(?<!\d))(\d+)?")  # g1: flag name g2: ephem?
FLAG_ARG_RE = re.compile(r"-+([a-zA-Z]\S*(?<!\d))(\d+)?")  # g1: flag name g2: ephem?
SINGLE_ARG_EXCEPTIONS = {"-i", "-h", "-v"}


def argsplit(args: str):
    view = CustomStringView(args.strip())
    args_out = []
    while not view.eof:
        view.skip_ws()
        word = view.get_quoted_word()
        if word is not None:
            args_out.append(word)
    return args_out


# ==== argparse ====
class Argument:
    def __init__(self, name: str, value, pos: int):
        self.name = name
        self.value = value
        self.pos = pos

    def __repr__(self):
        return f"<{type(self).__name__} name={self.name!r} value={self.value!r} pos={self.pos}>"

    def __eq__(self, other):
        return self.name == other.name and self.value == other.value and self.pos == other.pos


class EphemeralArgument(Argument):
    def __init__(self, name: str, value, pos: int, uses: int):
        super().__init__(name, value, pos)
        self.uses = uses
        self.used = 0

    def has_remaining_uses(self):
        return self.used < self.uses

    def __repr__(self):
        return (
            f"<{type(self).__name__} name={self.name!r} value={self.value!r} pos={self.pos} uses={self.uses} "
            f"used={self.used}>"
        )

    def __eq__(self, other):
        return (
            self.name == other.name and self.value == other.value and self.pos == other.pos and self.uses == other.uses
        )


def _argparse_arg(name: str, ephem: str | None, value, idx: int, parse_ephem: bool) -> Argument:
    if ephem and parse_ephem:
        return EphemeralArgument(name=name, value=value, pos=idx, uses=int(ephem))
    elif ephem:
        return Argument(name=name + ephem, value=value, pos=idx)
    else:
        return Argument(name=name, value=value, pos=idx)


def _argparse_iterator(args: list[str], parse_ephem: bool) -> Iterator[Argument]:
    flag_arg_state = None  # state: name, ephem?
    idx = 0
    for idx, arg in enumerate(args):
        # prio: single arg exceptions, flag args, values, single args
        if arg in SINGLE_ARG_EXCEPTIONS:
            if flag_arg_state is not None:
                name, ephem = flag_arg_state
                yield _argparse_arg(name, ephem, True, idx - 1, parse_ephem)
                flag_arg_state = None
            yield Argument(name=arg.lstrip("-"), value=True, pos=idx)
        elif match := FLAG_ARG_RE.fullmatch(arg):
            if flag_arg_state is not None:
                name, ephem = flag_arg_state
                yield _argparse_arg(name, ephem, True, idx - 1, parse_ephem)
            flag_arg_state = match.group(1), match.group(2)
        elif flag_arg_state is not None:
            name, ephem = flag_arg_state
            yield _argparse_arg(name, ephem, arg, idx - 1, parse_ephem)
            flag_arg_state = None
        elif match := SINGLE_ARG_RE.fullmatch(arg):
            name = match.group(1)
            ephem = match.group(2)
            yield _argparse_arg(name, ephem, True, idx, parse_ephem)
        # else: the current element at the head is junk

    if flag_arg_state is not None:
        name, ephem = flag_arg_state
        yield _argparse_arg(name, ephem, True, idx, parse_ephem)


# --- main entrypoint ---
def argparse(args, character=None, splitter=argsplit, parse_ephem=True) -> "ParsedArguments":
    """
    Given an argument string, returns the parsed arguments using the argument nondeterministic finite automaton.
    If *character* is given, evaluates {}-style math inside the string before parsing.
    If the argument is a string, uses *splitter* to split the string into args.
    If *parse_ephem* is False, arguments like ``-d1`` are saved literally rather than as an ephemeral argument.
    """
    if isinstance(args, str):
        args = splitter(args)

    if character:
        from aliasing.evaluators import MathEvaluator

        evaluator = MathEvaluator.with_character(character)
        args = [evaluator.transformed_str(a) for a in args]

    parsed_args = list(_argparse_iterator(args, parse_ephem))
    return ParsedArguments(parsed_args)


class ParsedArguments:
    def __init__(self, args: list[Argument]):
        self._parsed = collections.defaultdict(lambda: [])
        for arg in args:
            self._parsed[arg.name].append(arg)
        self._current_context = None
        self._contexts = {}  # type: dict[..., ParsedArguments]

    @classmethod
    def from_dict(cls, d):
        inst = cls([])
        for key, value in d.items():
            inst[key] = value
        return inst

    @classmethod
    def empty_args(cls):
        return cls([])

    def get(self, arg, default=None, type_=str, ephem=False):
        if default is None:
            default = []
        parsed = list(self._get_values(arg, ephem=ephem))
        if not parsed:
            return default
        try:
            return [type_(v) for v in parsed]
        except (ValueError, TypeError):
            raise InvalidArgument(f"One or more arguments cannot be cast to {type_.__name__} (in `{arg}`)")

    def last(self, arg, default=None, type_=str, ephem=False):
        last_arg = self._get_last(arg, ephem=ephem)
        if last_arg is None:
            return default
        try:
            return type_(last_arg)
        except (ValueError, TypeError):
            raise InvalidArgument(f"{last_arg} cannot be cast to {type_.__name__} (in `{arg}`)")

    def adv(self, eadv=False, boolwise=False, ephem=False, custom: dict = None):
        adv_str, dis_str, ea_str = "adv", "dis", "eadv"
        if custom is not None:
            if "adv" in custom:
                adv_str = custom["adv"]
            if "dis" in custom:
                dis_str = custom["dis"]
            if "eadv" in custom:
                ea_str = custom["eadv"]

        adv_arg = self.last(adv_str, default=False, type_=bool, ephem=ephem)
        dis_arg = self.last(dis_str, default=False, type_=bool, ephem=ephem)
        ea_arg = eadv and self.last(ea_str, default=False, type_=bool, ephem=ephem)

        if ea_arg and not dis_arg:
            out = 2
        elif dis_arg and not (adv_arg or ea_arg):
            out = -1
        elif adv_arg and not dis_arg:
            out = 1
        else:
            out = 0

        if not boolwise:
            return out
        else:
            return {-1: False, 0: None, 1: True}.get(out)

    def join(self, arg, connector: str, default=None, ephem=False):
        return connector.join(self.get(arg, ephem=ephem)) or default

    def ignore(self, arg):
        del self[arg]
        for context in self._contexts.values():
            del context[arg]

    def update(self, new):
        for k, v in new.items():
            self[k] = v

    def update_nx(self, new):
        for k, v in new.items():
            if k not in self and v is not None:
                self[k] = v

    @staticmethod
    def _yield_from_iterable(iterable: Iterator[Argument], ephem: bool):
        for value in iterable:
            if not ephem and isinstance(value, EphemeralArgument):
                continue
            elif isinstance(value, EphemeralArgument):
                if not value.has_remaining_uses():
                    continue
                value.used += 1
            yield value.value

    def _get_values(self, arg, ephem=False):
        iterable = self._parsed[arg]
        if self._current_context in self._contexts:
            iterable = itertools.chain(self._parsed[arg], self._contexts[self._current_context]._parsed[arg])
        yield from self._yield_from_iterable(iterable, ephem)

    def _get_last(self, arg, ephem=False):
        iterable = reversed(self._parsed[arg])
        if self._current_context in self._contexts:
            iterable = itertools.chain(
                reversed(self._contexts[self._current_context]._parsed[arg]), reversed(self._parsed[arg])
            )
        return next((self._yield_from_iterable(iterable, ephem)), None)

    def set_context(self, context):
        self._current_context = context

    def add_context(self, context, args):
        if isinstance(args, dict):
            if all(
                isinstance(k, (collections.UserString, str))
                and isinstance(v, (collections.UserList, list))
                and all(isinstance(i, (collections.UserString, str)) for i in v)
                for k, v in args.items()
            ):
                args = ParsedArguments.from_dict(args)
            else:
                raise InvalidArgument(f"Argument is not in the format dict[str, list[str]] (in {args})")
        elif not isinstance(args, ParsedArguments):
            raise InvalidArgument(f"Argument is not a dict or ParsedArguments (in {args})")

        self._contexts[context] = args

    def __contains__(self, item):
        return item in self._parsed and self._parsed[item]

    def __len__(self):
        return len(self._parsed)

    def __setitem__(self, key, value):
        if isinstance(value, (collections.UserList, list)):
            true_val = [Argument(key, v, idx) for idx, v in enumerate(value)]
        else:
            true_val = [Argument(key, value, 0)]
        self._parsed[key] = true_val
        if match := EPHEMERAL_ARG_RE.fullmatch(key):
            arg, num = match.group(1), match.group(2)
            ephem_val = [EphemeralArgument(arg, v.value, v.pos, int(num)) for v in true_val]
            self._parsed[arg] = ephem_val

    def __delitem__(self, arg):
        if arg in self._parsed:
            del self._parsed[arg]

    def __iter__(self):
        return iter(self._parsed.keys())

    def __repr__(self):
        return f"<ParsedArguments parsed={dict.__repr__(self._parsed)} context={self._current_context}>"


# ==== other helpers ====
def argquote(arg: str):
    if any(char in arg for char in string.whitespace):
        arg = arg.replace('"', '\\"')
        arg = f'"{arg}"'
    return arg


QUOTE_PAIRS = {
    '"': '"',
    "'": "'",
    "‘": "’",
    "‚": "‛",
    "“": "”",
    "„": "‟",
    "⹂": "⹂",
    "「": "」",
    "『": "』",
    "〝": "〞",
    "﹁": "﹂",
    "﹃": "﹄",
    "＂": "＂",
    "｢": "｣",
    "«": "»",
    "‹": "›",
    "《": "》",
    "〈": "〉",
}
ALL_QUOTES = set(QUOTE_PAIRS.keys()) | set(QUOTE_PAIRS.values())


class CustomStringView(StringView):
    def get_quoted_word(self):
        current = self.current
        if current is None:
            return None

        close_quote = QUOTE_PAIRS.get(current)
        is_quoted = bool(close_quote)
        if is_quoted:
            result = []
            _escaped_quotes = (current, close_quote)
        else:
            result = [current]
            _escaped_quotes = ALL_QUOTES

        while not self.eof:
            current = self.get()
            if not current:
                if is_quoted:
                    raise ExpectedClosingQuoteError(close_quote)
                return "".join(result)

            if current == "\\":
                next_char = self.get()
                if next_char in _escaped_quotes:
                    result.append(next_char)
                else:
                    self.undo()
                    result.append(current)
                continue

            if not is_quoted and current in ALL_QUOTES and current not in {"'", "’"}:
                close_quote = QUOTE_PAIRS.get(current)
                is_quoted = True
                _escaped_quotes = (current, close_quote)
                continue

            if is_quoted and current == close_quote:
                next_char = self.get()
                valid_eof = not next_char or next_char.isspace()
                if not valid_eof:
                    self.undo()
                    close_quote = None
                    is_quoted = False
                    _escaped_quotes = ALL_QUOTES
                    continue
                return "".join(result)

            if current.isspace() and not is_quoted:
                return "".join(result)

            result.append(current)


if __name__ == "__main__":  # pragma: no cover
    while True:
        try:
            print(argsplit(input(">>> ")))
        except BadArgument as e:
            print(e)



from __future__ import annotations

import ast
import logging
import inspect
from typing import Iterable, List, Sequence, Set

import draconic
from lsprotocol import types

from .argument_parsing import apply_argument_parsing
from .config import DiagnosticSettings
from .context import ContextData, GVarResolver
from .parser import find_draconic_blocks
from .runtime import MockExecutor, _default_builtins

log = logging.getLogger(__name__)

SEVERITY = {
    "error": types.DiagnosticSeverity.Error,
    "warning": types.DiagnosticSeverity.Warning,
    "info": types.DiagnosticSeverity.Information,
}


class DiagnosticProvider:
    def __init__(self, executor: MockExecutor, settings: DiagnosticSettings):
        self._executor = executor
        self._settings = settings
        self._builtin_signatures = _build_builtin_signatures()

    async def analyze(
        self,
        source: str,
        ctx_data: ContextData,
        gvar_resolver: GVarResolver,
    ) -> List[types.Diagnostic]:
        diagnostics: list[types.Diagnostic] = []

        source = apply_argument_parsing(source)
        blocks = find_draconic_blocks(source)
        if not blocks:
            diagnostics.extend(await self._analyze_code(source, ctx_data, gvar_resolver))
            return diagnostics

        for block in blocks:
            block_diags = await self._analyze_code(block.code, ctx_data, gvar_resolver)
            diagnostics.extend(_shift_diagnostics(block_diags, block.line_offset, block.char_offset))
        return diagnostics

    async def _analyze_code(
        self,
        code: str,
        ctx_data: ContextData,
        gvar_resolver: GVarResolver,
    ) -> List[types.Diagnostic]:
        diagnostics: list[types.Diagnostic] = []
        parser = draconic.DraconicInterpreter()
        line_shift = 0
        try:
            body = parser.parse(code)
        except draconic.DraconicSyntaxError as exc:
            wrapped, added = _wrap_draconic(code)
            try:
                body = parser.parse(wrapped)
                line_shift = -added
            except draconic.DraconicSyntaxError:
                diagnostics.append(_syntax_diagnostic(exc))
                return diagnostics
        except SyntaxError as exc:
            diagnostics.append(_syntax_from_std(exc))
            return diagnostics

        diagnostics.extend(
            self._check_unknown_names(body, ctx_data, self._settings.semantic_level)
        )
        diagnostics.extend(await _check_gvars(body, gvar_resolver, self._settings))
        diagnostics.extend(_check_imports(body, self._settings.semantic_level))
        diagnostics.extend(_check_call_args(body, self._builtin_signatures, self._settings.semantic_level))
        diagnostics.extend(_check_private_method_calls(body))
        if line_shift:
            diagnostics = _shift_diagnostics(diagnostics, line_shift, 0)
        return diagnostics

    def _check_unknown_names(
        self,
        body: Sequence[ast.AST],
        ctx_data: ContextData,
        severity_level: str,
    ) -> List[types.Diagnostic]:
        known: Set[str] = set(self._executor.available_names(ctx_data))
        diagnostics: list[types.Diagnostic] = []

        class Walker(ast.NodeVisitor):
            def __init__(self, tracker: Set[str]):
                self.tracker = tracker

            def visit_Assign(self, node: ast.Assign):
                self.visit(node.value)
                for target in node.targets:
                    self.tracker.update(_names_in_target(target))

            def visit_AnnAssign(self, node: ast.AnnAssign):
                if node.value:
                    self.visit(node.value)
                self.tracker.update(_names_in_target(node.target))

            def visit_AugAssign(self, node: ast.AugAssign):
                self.visit(node.value)
                self.tracker.update(_names_in_target(node.target))

            def visit_FunctionDef(self, node: ast.FunctionDef):
                self.tracker.add(node.name)
                for arg in node.args.args:
                    self.tracker.add(arg.arg)
                for stmt in node.body:
                    self.visit(stmt)

            def visit_ClassDef(self, node: ast.ClassDef):
                self.tracker.add(node.name)
                for stmt in node.body:
                    self.visit(stmt)

            def visit_For(self, node: ast.For):
                # Loop targets become defined for the loop body and after the loop
                self.tracker.update(_names_in_target(node.target))
                self.visit(node.iter)
                for stmt in node.body:
                    self.visit(stmt)
                for stmt in node.orelse:
                    self.visit(stmt)

            def visit_AsyncFor(self, node: ast.AsyncFor):
                # Async loop targets follow the same scoping rules as regular loops
                self.tracker.update(_names_in_target(node.target))
                self.visit(node.iter)
                for stmt in node.body:
                    self.visit(stmt)
                for stmt in node.orelse:
                    self.visit(stmt)

            def visit_Name(self, node: ast.Name):
                if isinstance(node.ctx, ast.Load) and node.id not in self.tracker:
                    diagnostics.append(
                        _make_diagnostic(
                            node,
                            f"'{node.id}' may be undefined in this scope",
                            severity_level,
                        )
                    )

        walker = Walker(known)
        for stmt in body:
            walker.visit(stmt)
        return diagnostics


def _syntax_diagnostic(exc: draconic.DraconicSyntaxError) -> types.Diagnostic:
    rng = _range_from_positions(
        exc.lineno,
        exc.offset,
        exc.end_lineno,
        exc.end_offset,
    )
    return types.Diagnostic(
        message=exc.msg,
        range=rng,
        severity=types.DiagnosticSeverity.Error,
        source="avrae-ls",
    )


def _syntax_from_std(exc: SyntaxError) -> types.Diagnostic:
    lineno, offset = exc.lineno, exc.offset
    rng = _range_from_positions(lineno, offset, getattr(exc, "end_lineno", None), getattr(exc, "end_offset", None))
    return types.Diagnostic(
        message=exc.msg,
        range=rng,
        severity=types.DiagnosticSeverity.Error,
        source="avrae-ls",
    )


def _names_in_target(target: ast.AST) -> Set[str]:
    names: set[str] = set()
    if isinstance(target, ast.Name):
        names.add(target.id)
    elif isinstance(target, ast.Tuple):
        for elt in target.elts:
            names.update(_names_in_target(elt))
    elif isinstance(target, ast.List):
        for elt in target.elts:
            names.update(_names_in_target(elt))
    return names


async def _check_gvars(
    body: Sequence[ast.AST],
    resolver: GVarResolver,
    settings: DiagnosticSettings,
) -> List[types.Diagnostic]:
    diagnostics: list[types.Diagnostic] = []
    for node in _iter_calls(body):
        if not isinstance(node.func, ast.Name) or node.func.id != "get_gvar":
            continue
        if not node.args:
            continue
        arg = node.args[0]
        if isinstance(arg, (ast.Str, ast.Constant)) and isinstance(getattr(arg, "s", None) or arg.value, str):
            gvar_id = arg.s if isinstance(arg, ast.Str) else arg.value
            found_local = resolver.get_local(gvar_id)
            ensured = found_local is not None or await resolver.ensure(gvar_id)
            if not ensured:
                diagnostics.append(
                    _make_diagnostic(
                        arg,
                        f"Unknown gvar '{gvar_id}'",
                        settings.semantic_level,
                    )
                )
    return diagnostics


def _iter_calls(body: Sequence[ast.AST]) -> Iterable[ast.Call]:
    class Finder(ast.NodeVisitor):
        def __init__(self):
            self.calls: list[ast.Call] = []

        def visit_Call(self, node: ast.Call):
            self.calls.append(node)
            self.generic_visit(node)

    finder = Finder()
    for stmt in body:
        finder.visit(stmt)
    return finder.calls


def _check_private_method_calls(body: Sequence[ast.AST]) -> List[types.Diagnostic]:
    diagnostics: list[types.Diagnostic] = []

    class Finder(ast.NodeVisitor):
        def visit_Call(self, node: ast.Call):
            func = node.func
            if isinstance(func, ast.Attribute) and func.attr.startswith("_"):
                diagnostics.append(
                    _make_diagnostic(
                        func,
                        "Calling private methods (starting with '_') is not allowed",
                        "error",
                    )
                )
            self.generic_visit(node)

    finder = Finder()
    for stmt in body:
        finder.visit(stmt)
    return diagnostics


def _make_diagnostic(node: ast.AST, message: str, level: str) -> types.Diagnostic:
    severity = SEVERITY.get(level, types.DiagnosticSeverity.Warning)
    if hasattr(node, "lineno"):
        rng = _range_from_positions(
            getattr(node, "lineno", 1),
            getattr(node, "col_offset", 0) + 1,
            getattr(node, "end_lineno", None),
            getattr(node, "end_col_offset", None),
        )
    else:
        rng = types.Range(
            start=types.Position(line=0, character=0),
            end=types.Position(line=0, character=1),
    )
    return types.Diagnostic(
        message=message,
        range=rng,
        severity=severity,
        source="avrae-ls",
    )


def _shift_diagnostics(diags: List[types.Diagnostic], line_offset: int, char_offset: int) -> List[types.Diagnostic]:
    shifted: list[types.Diagnostic] = []
    for diag in diags:
        shifted.append(
            types.Diagnostic(
                message=diag.message,
                range=_shift_range(diag.range, line_offset, char_offset),
                severity=diag.severity,
                source=diag.source,
                code=diag.code,
                code_description=diag.code_description,
                tags=diag.tags,
                related_information=diag.related_information,
                data=diag.data,
            )
        )
    return shifted


def _shift_range(rng: types.Range, line_offset: int, char_offset: int) -> types.Range:
    def _shift_pos(pos: types.Position) -> types.Position:
        return types.Position(
            line=max(pos.line + line_offset, 0),
            character=max(pos.character + (char_offset if pos.line == 0 else 0), 0),
        )

    return types.Range(start=_shift_pos(rng.start), end=_shift_pos(rng.end))


def _wrap_draconic(code: str) -> tuple[str, int]:
    indented = "\n".join(f"    {line}" for line in code.splitlines())
    wrapped = f"def __alias_main__():\n{indented}\n__alias_main__()"
    return wrapped, 1


def _build_builtin_signatures() -> dict[str, inspect.Signature]:
    sigs: dict[str, inspect.Signature] = {}
    builtins = _default_builtins()

    def try_add(name: str, obj):
        try:
            sigs[name] = inspect.signature(obj)
        except (TypeError, ValueError):
            pass

    for name, obj in builtins.items():
        try_add(name, obj)

    # runtime helpers we expose
    def get_gvar(key): ...
    def get_svar(name, default=None): ...
    def get_cvar(name, default=None): ...
    def get_uvar(name, default=None): ...
    def get_uvars(): ...
    def set_uvar(name, value): ...
    def set_uvar_nx(name, value): ...
    def delete_uvar(name): ...
    def uvar_exists(name): ...
    def exists(name): ...
    def get(name, default=None): ...
    def using(**imports): ...
    def signature(data=0): ...
    def verify_signature(sig=None): ...
    def print_fn(*args, sep=" ", end="\n"): ...

    helpers = {
        "get_gvar": get_gvar,
        "get_svar": get_svar,
        "get_cvar": get_cvar,
        "get_uvar": get_uvar,
        "get_uvars": get_uvars,
        "set_uvar": set_uvar,
        "set_uvar_nx": set_uvar_nx,
        "delete_uvar": delete_uvar,
        "uvar_exists": uvar_exists,
        "exists": exists,
        "get": get,
        "using": using,
        "signature": signature,
        "verify_signature": verify_signature,
        "print": print_fn,
    }
    for name, obj in helpers.items():
        try_add(name, obj)
    return sigs


def _check_call_args(
    body: Sequence[ast.AST],
    signatures: dict[str, inspect.Signature],
    severity_level: str,
) -> List[types.Diagnostic]:
    diagnostics: list[types.Diagnostic] = []

    class Visitor(ast.NodeVisitor):
        def visit_Call(self, node: ast.Call):
            if isinstance(node.func, ast.Name):
                fn = node.func.id
                if fn in signatures:
                    sig = signatures[fn]
                    if not _call_args_match(sig, node):
                        diagnostics.append(
                            _make_diagnostic(
                                node.func,
                                f"Call to '{fn}' may have invalid arguments",
                                severity_level,
                            )
                        )
            self.generic_visit(node)

    visitor = Visitor()
    for stmt in body:
        visitor.visit(stmt)
    return diagnostics


def _call_args_match(sig: inspect.Signature, call: ast.Call) -> bool:
    params = list(sig.parameters.values())
    required = [
        p
        for p in params
        if p.default is inspect._empty
        and p.kind
        in (
            inspect.Parameter.POSITIONAL_ONLY,
            inspect.Parameter.POSITIONAL_OR_KEYWORD,
        )
    ]
    max_args = None
    if any(p.kind == inspect.Parameter.VAR_POSITIONAL for p in params):
        max_args = None
    else:
        max_args = len(
            [
                p
                for p in params
                if p.kind
                in (
                    inspect.Parameter.POSITIONAL_ONLY,
                    inspect.Parameter.POSITIONAL_OR_KEYWORD,
                )
            ]
        )

    arg_count = len(call.args)
    if arg_count < len(required):
        return False
    if max_args is not None and arg_count > max_args:
        return False
    return True


def _check_imports(body: Sequence[ast.AST], severity_level: str) -> List[types.Diagnostic]:
    diagnostics: list[types.Diagnostic] = []

    class Visitor(ast.NodeVisitor):
        def visit_Import(self, node: ast.Import):
            diagnostics.append(_make_diagnostic(node, "Imports are not supported in draconic aliases", severity_level))

        def visit_ImportFrom(self, node: ast.ImportFrom):
            diagnostics.append(_make_diagnostic(node, "Imports are not supported in draconic aliases", severity_level))

    visitor = Visitor()
    for stmt in body:
        visitor.visit(stmt)
    return diagnostics


def _range_from_positions(
    lineno: int | None,
    col_offset: int | None,
    end_lineno: int | None,
    end_col_offset: int | None,
) -> types.Range:
    start = types.Position(
        line=max((lineno or 1) - 1, 0),
        character=max((col_offset or 1) - 1, 0),
    )
    end = types.Position(
        line=max(((end_lineno or lineno or 1) - 1), 0),
        character=max(((end_col_offset or col_offset or 1) - 1), 0),
    )
    return types.Range(start=start, end=end)



import abc
import ast
import contextlib
from collections.abc import Mapping, Sequence
from functools import cached_property

from .exceptions import *
from .helpers import DraconicConfig, OperatorMixin, zip_star
from .string import check_format_spec
from .versions import PY_310
from .types import approx_len_of

__all__ = ("SimpleInterpreter", "DraconicInterpreter")


# ===== single-line evaluator, noncompound types, etc =====
class SimpleInterpreter(OperatorMixin):
    """A simple interpreter capable of evaluating expressions. No compound types or assignments."""

    def __init__(self, builtins=None, config=None):
        if config is None:
            config = DraconicConfig()
        if builtins is None:
            builtins = {}

        super().__init__(config)

        if config.builtins_extend_default:
            builtins = {**config.default_names, **builtins}

        self.builtins = builtins

        self.nodes = {
            ast.Expr: self._eval_expr,
            # literals:
            ast.Num: self._eval_num,
            ast.Str: self._eval_str,
            ast.Constant: self._eval_constant,
            ast.FormattedValue: self._eval_formattedvalue,  # formatted value in f-string
            ast.JoinedStr: self._eval_joinedstr,  # f-string
            ast.NameConstant: self._eval_constant,  # True/False/None up to py3.7
            # names:
            ast.Name: self._eval_name,
            # ops:
            ast.UnaryOp: self._eval_unaryop,
            ast.BinOp: self._eval_binop,
            ast.BoolOp: self._eval_boolop,
            ast.Compare: self._eval_compare,
            ast.IfExp: self._eval_ifexp,
            # function call:
            ast.Call: self._eval_call,
            ast.keyword: self._eval_keyword,  # foo(x=y), kwargs (not supported)
            # container[key]:
            ast.Subscript: self._eval_subscript,
            ast.Index: self._eval_index,  # deprecated in py3.9 (bpo-34822)
            ast.Slice: self._eval_slice,  # deprecated in py3.9 (bpo-34822)
            # container.key:
            ast.Attribute: self._eval_attribute,
        }

        self._str = self._config.str
        self._expr = None  # save the expression for error handling

    def parse(self, expr: str):
        """
        Parses an expression.

        :type expr: str
        :rtype: list[ast.AST]
        """
        self._expr = expr
        try:
            return ast.parse(expr).body
        except SyntaxError as e:
            raise DraconicSyntaxError(e, expr) from e

    def eval(self, expr: str):
        """
        Evaluates an expression.

        :type expr: list[ast.AST] or str
        """
        expr = self.parse(expr)

        self._preflight()
        try:
            expression = expr[0]
        except IndexError:  # if there is no expression, evaluate to None
            return None
        return self._eval(expression)

    def _eval(self, node):
        """The internal evaluator used on each node in the parsed tree."""
        try:
            handler = self.nodes[type(node)]
        except KeyError:
            raise FeatureNotAvailable(
                "Sorry, {0} is not available in this evaluator".format(type(node).__name__), node, self._expr
            )

        try:
            return handler(node)
        except _PostponedRaise as pr:
            raise pr.cls(*pr.args, **pr.kwargs, node=node, expr=self._expr)
        except DraconicException:
            raise
        except Exception as e:
            raise AnnotatedException(e, node, self._expr) from e

    def _preflight(self):
        """Called before starting evaluation."""
        pass

    @property
    def names(self):
        return self.builtins

    @names.setter
    def names(self, new_names):
        self.builtins = new_names

    # ===== nodes =====
    def _eval_expr(self, node):
        return self._eval(node.value)

    @staticmethod
    def _eval_num(node):
        return node.n

    def _eval_str(self, node):
        if len(node.s) > self._config.max_const_len:
            raise IterableTooLong(
                f"String literal in statement is too long ({len(node.s)} > {self._config.max_const_len})",
                node,
                self._expr,
            )
        return self._str(node.s)

    def _eval_constant(self, node):
        if hasattr(node.value, "__len__") and len(node.value) > self._config.max_const_len:
            raise IterableTooLong(
                f"Literal in statement is too long ({len(node.value)} > {self._config.max_const_len})", node, self._expr
            )
        if isinstance(node.value, bytes):
            raise FeatureNotAvailable("Creation of bytes literals is not allowed", node, self._expr)
        return node.value

    def _eval_unaryop(self, node):
        return self.operators[type(node.op)](self._eval(node.operand))

    def _eval_binop(self, node):
        return self.operators[type(node.op)](self._eval(node.left), self._eval(node.right))

    def _eval_boolop(self, node):
        vout = False
        if isinstance(node.op, ast.And):
            for value in node.values:
                vout = self._eval(value)
                if not vout:
                    return vout
        elif isinstance(node.op, ast.Or):
            for value in node.values:
                vout = self._eval(value)
                if vout:
                    return vout
        return vout

    def _eval_compare(self, node):
        right = self._eval(node.left)
        to_return = True
        for operation, comp in zip(node.ops, node.comparators):
            if not to_return:
                break
            left = right
            right = self._eval(comp)
            to_return = self.operators[type(operation)](left, right)
        return to_return

    def _eval_ifexp(self, node):
        return self._eval(node.body) if self._eval(node.test) else self._eval(node.orelse)

    def _eval_call(self, node):
        func = self._eval(node.func)
        return func(*(self._eval(a) for a in node.args), **dict(self._eval(k) for k in node.keywords))

    def _eval_keyword(self, node):
        return node.arg, self._eval(node.value)

    def _eval_name(self, node):
        try:
            return self.names[node.id]
        except KeyError:
            raise NotDefined(f"{node.id} is not defined", node, self._expr)

    def _eval_subscript(self, node):
        container = self._eval(node.value)
        key = self._eval(node.slice)
        try:
            return container[key]
        except KeyError:
            raise

    def _eval_attribute(self, node):
        for prefix in self._config.disallow_prefixes:
            if node.attr.startswith(prefix):
                raise FeatureNotAvailable(f"Access to the {node.attr} attribute is not allowed", node, self._expr)
        if node.attr in self._config.disallow_methods:
            raise FeatureNotAvailable(f"Access to the {node.attr} attribute is not allowed", node, self._expr)
        # eval node
        node_evaluated = self._eval(node.value)

        # Maybe the base object is an actual object, not just a dict
        try:
            return getattr(node_evaluated, node.attr)
        except (AttributeError, TypeError):
            # If it is not present, raise an exception
            raise NotDefined(f"'{type(node_evaluated).__name__}' object has no attribute {node.attr}", node, self._expr)

    def _eval_index(self, node):
        return self._eval(node.value)

    def _eval_slice(self, node):
        lower = upper = step = None
        if node.lower is not None:
            lower = self._eval(node.lower)
        if node.upper is not None:
            upper = self._eval(node.upper)
        if node.step is not None:
            step = self._eval(node.step)
        return slice(lower, upper, step)

    def _eval_joinedstr(self, node):
        length = 0
        evaluated_values = []
        for n in node.values:
            val = str(self._eval(n))
            length += len(val)
            if length > self._config.max_const_len:
                raise IterableTooLong(
                    f"f-string in statement is too long ({length} > {self._config.max_const_len})", node, self._expr
                )
            evaluated_values.append(val)
        return "".join(evaluated_values)

    def _eval_formattedvalue(self, node):
        if node.format_spec:
            format_spec = str(self._eval(node.format_spec))
            check_format_spec(self._config, format_spec)
            return self._str(format(self._eval(node.value), format_spec))
        return self._eval(node.value)


# ===== multiple-line execution, assignment, compound types =====
class _Break:
    __slots__ = ("node",)

    def __init__(self, node: ast.Break):
        self.node = node


class _Continue:
    __slots__ = ("node",)

    def __init__(self, node: ast.Continue):
        self.node = node


class _Return:
    __slots__ = ("value", "node")

    def __init__(self, retval, node: ast.Return):
        self.value = retval
        self.node = node


class _Callable(abc.ABC):
    """ABC for functions and lambdas"""

    def __init__(self, interpreter, node, names_at_def, defining_expr):
        self._interpreter = interpreter
        self._node = node
        self._outer_scope_names = names_at_def
        self._defining_expr = defining_expr

    # faux introspection props since __dunders__ aren't accessible
    @property
    def name(self):
        return self.__name__

    @cached_property
    def doc(self):
        return ast.get_docstring(self._node)

    def __call__(self, *args, **kwargs):
        raise NotImplementedError


class _Function(_Callable):
    """A wrapper class around an ast.FunctionDef."""

    def __init__(self, interpreter, functiondef, names_at_def, defining_expr):
        super().__init__(interpreter, functiondef, names_at_def, defining_expr)
        self.__name__ = self._name = functiondef.name

    def __repr__(self):
        return f"<Function {self._name}>"

    def __call__(self, *args, **kwargs):
        try:
            # noinspection PyProtectedMember
            return self._interpreter._exec_function(self, *args, **kwargs)
        except DraconicException as e:
            e.__drac_context__ = self._name
            raise


class _Lambda(_Callable):
    """A wrapper class around an ast.Lambda."""

    def __init__(self, interpreter, lambdadef, names_at_def, defining_expr):
        super().__init__(interpreter, lambdadef, names_at_def, defining_expr)
        self.__name__ = self._name = "<lambda>"

    def __repr__(self):
        return f"<Function <lambda>>"

    @property
    def doc(self):
        return None

    def __call__(self, *args, **kwargs):
        try:
            # noinspection PyProtectedMember
            return self._interpreter._exec_lambda(self, *args, **kwargs)
        except DraconicException as e:
            e.__drac_context__ = self._name
            raise


class DraconicInterpreter(SimpleInterpreter):
    """The Draconic interpreter. Capable of running Draconic code."""

    def __init__(self, builtins=None, config=None, initial_names=None):
        super().__init__(builtins, config)

        if initial_names is None:
            initial_names = {}

        self.nodes.update(
            {
                # compound types:
                ast.Dict: self._eval_dict,
                ast.Tuple: self._eval_tuple,
                ast.List: self._eval_list,
                ast.Set: self._eval_set,
                # comprehensions:
                ast.ListComp: self._eval_listcomp,
                ast.SetComp: self._eval_setcomp,
                ast.DictComp: self._eval_dictcomp,
                ast.GeneratorExp: self._eval_generatorexp,
                ast.Starred: self._eval_starred,  # foo(*iterable), [*iterable], etc.
                # assignments:
                ast.Assign: self._eval_assign,
                ast.AugAssign: self._eval_augassign,
                ast.NamedExpr: self._eval_namedexpr,
                # control:
                ast.Return: self._exec_return,
                ast.If: self._exec_if,
                ast.For: self._exec_for,
                ast.While: self._exec_while,
                ast.Break: lambda node: _Break(node),
                ast.Continue: lambda node: _Continue(node),
                ast.Pass: lambda node: None,
                # functions:
                ast.FunctionDef: self._eval_functiondef,
                ast.Lambda: self._eval_lambda,
                # try/except:
                ast.Try: self._exec_try,
            }
        )

        self.assign_nodes = {
            ast.Name: self._assign_name,
            ast.Tuple: self._assign_unpack,
            ast.List: self._assign_unpack,
            ast.Subscript: self._assign_subscript,
            # no assigning to attributes
            ast.Starred: self._assign_starred,  # a, *b = [x, y, z]
        }

        self.patma_nodes = {}

        if PY_310:
            self.nodes.update(
                {
                    ast.Match: self._exec_match,
                }
            )

            self.patma_nodes.update(
                {
                    ast.MatchValue: self._patma_match_value,
                    ast.MatchSingleton: self._patma_match_singleton,
                    ast.MatchSequence: self._patma_match_sequence,
                    ast.MatchMapping: self._patma_match_mapping,
                    ast.MatchStar: self._patma_match_star,
                    # no MatchClass
                    ast.MatchAs: self._patma_match_as,
                    ast.MatchOr: self._patma_match_or,
                }
            )

        # compound type helpers
        self._list = self._config.list
        self._set = self._config.set
        self._dict = self._config.dict

        self._num_stmts = 0
        self._loops = 0
        self._depth = 1
        self._names = initial_names

    def eval(self, expr: str):
        retval = super().eval(expr)
        if isinstance(retval, _Return):
            return retval.value
        elif isinstance(retval, (_Break, _Continue)):
            raise DraconicSyntaxError.from_node(retval.node, msg="Loop control outside loop", expr=self._expr)
        return retval

    def execute(self, expr: str):
        """Executes an AST body."""
        expr = self.parse(expr)

        self._preflight()
        retval = self._exec(expr)
        if isinstance(retval, (_Break, _Continue)):
            raise DraconicSyntaxError.from_node(retval.node, msg="Loop control outside loop", expr=self._expr)
        if isinstance(retval, _Return):
            return retval.value

    def execute_module(self, expr: str, module_name="<module>"):
        """
        Executes the expression as if it was a module.
        This is similar to *execute* except:
        - it doesn't allow bare returns
        - it doesn't call preflight
        - it saves the previously running expression
        - it sets the exception context
        """
        old_expr = self._expr
        try:
            expr = self.parse(expr)
            retval = self._exec(expr)
            if isinstance(retval, (_Break, _Continue)):
                raise DraconicSyntaxError.from_node(retval.node, msg="Loop control outside loop", expr=self._expr)
            if isinstance(retval, _Return):
                raise DraconicSyntaxError.from_node(retval.node, msg="'return' outside function", expr=self._expr)
        except DraconicException as e:
            e.__drac_context__ = module_name
            raise
        finally:
            self._expr = old_expr

    def _preflight(self):
        self._num_stmts = 0
        self._loops = 0
        super()._preflight()

    def _eval(self, node):
        self._num_stmts += 1
        if self._num_stmts > self._config.max_statements:
            raise TooManyStatements("You are trying to execute too many statements.", node, self._expr)

        val = super()._eval(node)
        # ensure that it's always an instance of our safe compound types being returned
        # note: makes a copy, so the original copy won't be updated
        # we don't use isinstance because we're looking for very specific classes
        if type(val) is str:
            return self._str(val)
        elif type(val) is list:
            return self._list(val)
        elif type(val) is dict:
            return self._dict(val)
        elif type(val) is set:
            return self._set(val)
        return val

    def _exec(self, body):
        for expression in body:
            retval = self._eval(expression)
            if isinstance(retval, (_Return, _Break, _Continue)):
                return retval

    @property
    def names(self):
        return {**self.builtins, **self._names}

    @names.setter
    def names(self, new_names):
        self._names = new_names

    # ===== compound types =====
    def _eval_dict(self, node):
        return self._dict(self._starred_keyword_unwrap(zip(node.keys, node.values)))

    def _eval_tuple(self, node):
        return tuple(self._starred_unwrap(node.elts))

    def _eval_list(self, node):
        return self._list(self._starred_unwrap(node.elts))

    def _eval_set(self, node):
        return self._set(self._starred_unwrap(node.elts))

    def _eval_listcomp(self, node):
        return self._list(self._do_comprehension(node))

    def _eval_setcomp(self, node):
        return self._set(self._do_comprehension(node))

    def _eval_dictcomp(self, node):
        return self._dict(self._do_comprehension(node, is_dictcomp=True))

    def _eval_generatorexp(self, node):
        for item in self._do_comprehension(node):
            yield item

    def _do_comprehension(self, comprehension_node, is_dictcomp=False):
        if is_dictcomp:

            def do_value(node):
                return self._eval(node.key), self._eval(node.value)

        else:

            def do_value(node):
                return self._eval(node.elt)

        extra_names = {}
        previous_name_evaller = self.nodes[ast.Name]

        def eval_names_extra(node):
            """
            Here we hide our extra scope for within this comprehension
            """
            if node.id in extra_names:
                return extra_names[node.id]
            return previous_name_evaller(node)

        self.nodes.update({ast.Name: eval_names_extra})

        def recurse_targets(target, value):
            """
                Recursively (enter, (into, (nested, name), unpacking)) = \
                             and, (assign, (values, to), each
            """
            if isinstance(target, ast.Name):
                extra_names[target.id] = value
            else:
                for t, v in zip(target.elts, value):
                    recurse_targets(t, v)

        def do_generator(gi=0, total_len=0):
            """
            For each generator, set the names used in the final emitted value/the next generator.
            Only the final generator (gi = len(comprehension_node.generator)-1) should emit the final values,
            since only then are all possible necessary values set in extra_names.
            """
            generator_node = comprehension_node.generators[gi]
            for i in self._eval(generator_node.iter):
                self._loops += 1
                if self._loops > self._config.max_loops:
                    raise IterableTooLong("Comprehension generates too many elements", comprehension_node, self._expr)

                # set names
                recurse_targets(generator_node.target, i)

                if all(self._eval(iff) for iff in generator_node.ifs):
                    if len(comprehension_node.generators) > gi + 1:
                        # next generator
                        yield from do_generator(gi + 1, total_len)  # bubble up emitted values
                    else:
                        # emit values
                        value = do_value(comprehension_node)
                        total_len += sum(approx_len_of(val) for val in value) if is_dictcomp else approx_len_of(value)
                        total_len += 1
                        if total_len > self._config.max_const_len:
                            raise IterableTooLong("Comprehension generates too much", comprehension_node, self._expr)
                        yield value

        try:
            yield from do_generator()
        finally:
            self.nodes.update({ast.Name: previous_name_evaller})

    def _eval_starred(self, node):
        raise DraconicSyntaxError.from_node(node, "can't use starred expression here", self._expr)

    def _starred_unwrap(self, nodes, *, check_len=True):
        total_len = 0

        for node in nodes:
            if type(node) is ast.Starred:
                evalue = self._eval(node.value)
                try:
                    for retval in evalue:
                        self._loops += 1
                        if self._loops > self._config.max_loops:
                            raise IterableTooLong("Unwrapping generates too many elements", node, self._expr)
                        if check_len:
                            total_len += approx_len_of(retval) + 1
                            if total_len > self._config.max_const_len:
                                raise IterableTooLong("Unwrapping generates too much", node, self._expr)
                        yield retval
                except TypeError:
                    raise TypeError(f"Value after * must be iterable, got {type(evalue).__name__}")
            else:
                retval = self._eval(node)
                if check_len:
                    total_len += approx_len_of(retval) + 1
                    if total_len > self._config.max_const_len:
                        raise IterableTooLong("Unwrapping generates too much", node, self._expr)
                yield retval

    def _starred_keyword_unwrap(self, items, *, check_len=True):
        total_len = 0

        for key, value in items:
            evalue = self._eval(value)
            if key is None:
                if isinstance(evalue, Mapping):
                    for retval in evalue.items():
                        self._loops += 1
                        if self._loops > self._config.max_loops:
                            raise IterableTooLong("Unwrapping generates too many elements", value, self._expr)
                        if check_len:
                            total_len += sum(approx_len_of(val) for val in retval) + 1
                            if total_len > self._config.max_const_len:
                                raise IterableTooLong("Unwrapping generates too much", value, self._expr)
                        yield retval
                else:
                    raise TypeError(f"argument after ** must be a mapping, got {type(value).__name__}")
            else:
                retval = self._eval(key) if isinstance(key, ast.AST) else key, evalue
                if check_len:
                    total_len += sum(approx_len_of(val) for val in retval) + 1
                    if total_len > self._config.max_const_len:
                        raise IterableTooLong("Unwrapping generates too much", value, self._expr)
                yield retval

    # ===== assignments =====
    def _eval_assign(self, node):
        value = self._eval(node.value)
        for target in node.targets:  # a = b = 1
            self._assign(target, value)

    def _eval_augassign(self, node):
        target = node.target
        # transform a += 1 to a = a + 1, then we can use assign and eval
        new_value = ast.BinOp(left=target, op=node.op, right=node.value)
        ast.copy_location(new_value, target)
        self._assign(target, self._eval_binop(new_value))

    def _eval_namedexpr(self, node):
        value = self._eval(node.value)
        self._assign(node.target, value)
        return value

    # ---- primary assign branch ----
    def _assign(self, names, values):
        try:
            handler = self.assign_nodes[type(names)]
        except KeyError:
            raise FeatureNotAvailable(f"Assignment to {type(names).__name__} is not allowed", names, self._expr)
        # noinspection PyArgumentList
        return handler(names, values)

    def _assign_name(self, name, value):
        if name.id in self.builtins:
            raise DraconicValueError(f"{name.id} is already builtin (no shadow assignments).", name, self._expr)
        self._names[name.id] = value

    def _assign_subscript(self, name, value):
        container = self._eval(name.value)
        key = self._eval(name.slice)
        container[key] = value  # no further evaluation needed, if container is in names it will update

    def _assign_unpack(self, names, values):
        if not isinstance(names, (ast.Tuple, ast.List)):
            self._assign(names, values)
        else:
            try:
                values = list(iter(values))
            except TypeError:
                raise DraconicValueError(
                    f"Cannot unpack non-iterable {type(values).__name__} object", names, self._expr
                )

            stars = (i for i in names.elts if type(i) is ast.Starred)
            starred = next(stars, None)

            if starred is None:
                if len(names.elts) > len(values):
                    raise DraconicValueError(
                        f"not enough values to unpack (expected {len(names.elts)}, got {len(values)})",
                        names,
                        self._expr,
                    )
                elif len(names.elts) < len(values):
                    raise DraconicValueError(
                        f"too many values to unpack (expected {len(names.elts)}, got {len(values)})",
                        names,
                        self._expr,
                    )
                for t, v in zip(names.elts, values):
                    self._assign_unpack(t, v)
            elif (extra := next(stars, None)) is not None:
                raise DraconicSyntaxError.from_node(extra, "multiple starred expressions in assignment", self._expr)
            else:
                if len(values) < (len(names.elts) - 1):
                    raise DraconicValueError(
                        f"not enough values to unpack (expected at least {len(names.elts) - 1}, got {len(values)})",
                        names,
                        self._expr,
                    )

                for t, v in zip_star(names.elts, values, star_index=names.elts.index(starred)):
                    self._assign_unpack(t, v)

    def _assign_starred(self, name, value):
        self._assign_name(name.value, value)

    # ===== execution =====
    def _exec_return(self, node):
        retval = self._eval(node.value) if node.value is not None else None
        return _Return(retval, node)

    def _exec_if(self, node):
        test = self._eval(node.test)
        if test:
            return self._exec(node.body)
        else:
            return self._exec(node.orelse)

    def _exec_for(self, node):
        for item in self._eval(node.iter):
            self._loops += 1
            if self._loops > self._config.max_loops:
                raise TooManyStatements("Too many loops (in for block)", node, self._expr)

            self._assign(node.target, item)
            retval = self._exec(node.body)
            if isinstance(retval, _Return):
                return retval
            elif isinstance(retval, _Break):
                break
            elif isinstance(retval, _Continue):
                continue
        else:
            return self._exec(node.orelse)

    def _exec_while(self, node):
        while self._eval(node.test):
            self._loops += 1
            if self._loops > self._config.max_loops:
                raise TooManyStatements("Too many loops (in while block)", node, self._expr)

            retval = self._exec(node.body)
            if isinstance(retval, _Return):
                return retval
            elif isinstance(retval, _Break):
                break
            elif isinstance(retval, _Continue):
                continue
        else:
            return self._exec(node.orelse)

    # ===== patma =====
    # impl inspired by GVR's impl at https://github.com/gvanrossum/patma/blob/master/patma.py
    # note: we do duplicate binding checks at runtime instead of preflight, which means that
    # some code could run before the duplicate binding is detected, and certain exprs illegal in Python are legal here
    # this is OK for our use case but differs from Python's impl
    def _exec_match(self, node):
        subject = self._eval(node.subject)
        for match_case in node.cases:
            if (bindings := self._patma(match_case.pattern, subject)) is not None:
                self._names.update(bindings)  # In python patma, values are bound before the guard executes
                if match_case.guard is not None and not self._eval(match_case.guard):
                    continue
                return self._exec(match_case.body)

    def _patma(self, pattern, subject):
        """
        Execute matching logic for a given match_case and subject.
        If the subject matches the case, return the dict of bindings for this case.
        Otherwise, return None.
        """
        try:
            handler = self.patma_nodes[type(pattern)]
        except KeyError:
            raise FeatureNotAvailable(f"Matching on {type(pattern).__name__} is not allowed", pattern, self._expr)
        self._num_stmts += 1
        return handler(pattern, subject)

    def _patma_match_value(self, node, subject):
        if subject == self._eval(node.value):
            return {}
        return None

    @staticmethod
    def _patma_match_singleton(node, subject):
        if subject is node.value:
            return {}
        return None

    def _patma_match_sequence(self, node, subject):
        if not isinstance(subject, Sequence) or isinstance(subject, (str, bytes)):
            return None

        match_star_idxs = [idx for idx, pattern in enumerate(node.patterns) if isinstance(pattern, ast.MatchStar)]
        if len(match_star_idxs) > 1:
            # multiple starred names
            raise DraconicValueError(f"multiple starred names in sequence pattern", node, self._expr)
        elif match_star_idxs:
            # one starred name
            if not len(node.patterns) <= len(subject) + 1:
                return None
            pattern_iterator = zip_star(node.patterns, subject, star_index=match_star_idxs[0])
        else:
            # no starred names
            if len(node.patterns) != len(subject):
                return None
            pattern_iterator = zip(node.patterns, subject)

        # do iteration over patterns and values
        bindings = {}
        bound_names = set()
        for pattern, item in pattern_iterator:
            # recursive check
            # noinspection DuplicatedCode
            match = self._patma(pattern, item)
            if match is None:
                return None

            # duplicate bindings check
            if bound_names.intersection(match):
                raise DraconicValueError(
                    f"multiple assignment to names {sorted(bound_names.intersection(match))} in sequence pattern",
                    node,
                    self._expr,
                )
            bindings.update(match)
            bound_names.update(match)

        return bindings

    def _patma_match_mapping(self, node, subject):
        if not isinstance(subject, Mapping):
            return None

        bindings = {}
        bound_names = set()
        bound_keys = set()
        for key, pattern in zip(node.keys, node.patterns):
            # recursive check
            key = self._eval(key)
            try:
                value = subject[key]
            except KeyError:
                return None
            # noinspection DuplicatedCode
            match = self._patma(pattern, value)
            if match is None:
                return None

            # duplicate bindings check
            if bound_names.intersection(match):
                raise DraconicValueError(
                    f"multiple assignment to names {sorted(bound_names.intersection(match))} in mapping pattern",
                    node,
                    self._expr,
                )
            bindings.update(match)
            bound_names.update(match)
            bound_keys.add(key)

        if node.rest is not None:
            if node.rest in bound_names:
                raise DraconicValueError(
                    f"multiple assignment to name {node.rest!r} in mapping pattern", node, self._expr
                )
            bindings[node.rest] = {k: v for k, v in subject.items() if k not in bound_keys}

        return bindings

    @staticmethod
    def _patma_match_star(node, subject):
        if node.name is None:
            return {}
        return {node.name: subject}

    def _patma_match_as(self, node, subject):
        if node.name is None:  # this is the wildcard pattern, always matches
            return {}
        if node.pattern is None:  # bare name capture pattern, always matches
            return {node.name: subject}
        # otherwise if the inner match matches, we just add an additional binding to it
        inner_match = self._patma(node.pattern, subject)
        if inner_match is None:
            return None
        return {**inner_match, node.name: subject}

    def _patma_match_or(self, node, subject):
        # since we don't know the subpattern's bindings until it executes, we can't enforce both sides having the
        # same bindings like in Python
        for pattern in node.patterns:
            match = self._patma(pattern, subject)
            if match is not None:
                return match
        return None

    # ===== functions =====
    # definitions
    def _eval_functiondef(self, node):
        if node.name in self.builtins:
            raise DraconicValueError(f"{node.name} is already builtin (no shadow assignments).", node, self._expr)
        self._names[node.name] = _Function(self, node, self._names, self._expr)

    def _eval_lambda(self, node):
        return _Lambda(self, node, self._names, self._expr)

    # executions
    def _eval_call(self, node):
        func = self._eval(node.func)
        args = tuple(self._starred_unwrap(node.args, check_len=False))
        kwargs = dict(self._starred_keyword_unwrap(((k.arg, k.value) for k in node.keywords), check_len=False))
        try:
            return func(*args, **kwargs)
        except DraconicException as e:
            raise NestedException(e.msg, node, self._expr, last_exc=e) from e

    # noinspection PyProtectedMember
    @contextlib.contextmanager
    def _function_call_context(self, __functiondef, /, *args, **kwargs):
        # check limits
        self._depth += 1
        if self._depth > self._config.max_recursion_depth:
            _raise_in_context(TooMuchRecursion, "Maximum recursion depth exceeded")
        # store current names and expression
        old_names = self._names
        old_expr = self._expr
        # bind closure names and contextual expression
        self._names = __functiondef._outer_scope_names.copy()
        self._expr = __functiondef._defining_expr

        # yield control to the call
        try:
            self._bind_function_args(__functiondef, *args, **kwargs)
            yield
        finally:
            # restore old names and expr
            self._names = old_names
            self._expr = old_expr
            # reduce recursion depth
            self._depth -= 1

    # noinspection PyProtectedMember
    def _bind_function_args(self, __functiondef, /, *args, **kwargs):
        # check and bind args
        arguments = __functiondef._node.args
        # check valid pos num
        if len(args) > (numpos := len(arguments.posonlyargs) + len(arguments.args)) and arguments.vararg is None:
            raise TypeError(f"{__functiondef._name}() takes {numpos} positional arguments but {len(args)} were given")
        args_i = 0
        default_i = len(arguments.defaults) - numpos
        # posonly
        for posonly in arguments.posonlyargs:
            if args_i + 1 > len(args):
                if default_i < 0:
                    raise TypeError(f"{__functiondef._name}() missing required positional argument: {posonly.arg!r}")
                self._names[posonly.arg] = self._eval(arguments.defaults[default_i])
            else:
                self._names[posonly.arg] = args[args_i]
            args_i += 1
            default_i += 1
        # normal
        for posarg in arguments.args:
            # at least 1
            if args_i + 1 > len(args) and posarg.arg not in kwargs:
                if default_i < 0:
                    raise TypeError(f"{__functiondef._name}() missing required positional argument: {posarg.arg!r}")
                self._names[posarg.arg] = self._eval(arguments.defaults[default_i])
            # pos and kw
            elif args_i + 1 <= len(args) and posarg.arg in kwargs:
                raise TypeError(f"{__functiondef._name}() got multiple values for argument {posarg.arg!r}")
            elif posarg.arg in kwargs:
                self._names[posarg.arg] = kwargs.pop(posarg.arg)
            else:
                # we won't indexerror because if it's not in kwargs and args_i is invalid, the first if catches it
                self._names[posarg.arg] = args[args_i]
            args_i += 1
            default_i += 1
        # kwargonly
        for k_i, kwargonly in enumerate(arguments.kwonlyargs):
            if kwargonly.arg not in kwargs and arguments.kw_defaults[k_i] is None:
                raise TypeError(f"{__functiondef._name}() missing required keyword argument: {kwargonly.arg!r}")
            if kwargonly.arg in kwargs:
                self._names[kwargonly.arg] = kwargs.pop(kwargonly.arg)
            else:
                self._names[kwargonly.arg] = self._eval(arguments.kw_defaults[k_i])
        # *args
        if arguments.vararg is not None:
            if approx_len_of(args[args_i:]) > self._config.max_const_len:
                _raise_in_context(IterableTooLong, f"*{arguments.vararg.arg} would be too large")
            self._names[arguments.vararg.arg] = tuple(args[args_i:])
        # **kwargs
        if arguments.kwarg is not None:
            if approx_len_of(kwargs) > self._config.max_const_len:
                _raise_in_context(IterableTooLong, f"**{arguments.kwarg.arg} would be too large")
            self._names[arguments.kwarg.arg] = kwargs
        elif kwargs:  # and arguments.kwarg is None (implicit)
            raise TypeError(f"{__functiondef._name}() got unexpected keyword arguments: {tuple(kwargs.keys())}")

    # noinspection PyProtectedMember
    def _exec_function(self, __functiondef: _Function, /, *args, **kwargs):
        with self._function_call_context(__functiondef, *args, **kwargs):
            retval = self._exec(__functiondef._node.body)
            if isinstance(retval, (_Break, _Continue)):
                raise DraconicSyntaxError.from_node(retval.node, msg="Loop control outside loop", expr=self._expr)
            if isinstance(retval, _Return):
                return retval.value

    # noinspection PyProtectedMember
    def _exec_lambda(self, __lambdadef: _Lambda, /, *args, **kwargs):
        with self._function_call_context(__lambdadef, *args, **kwargs):
            return self._eval(__lambdadef._node.body)

    # ===== try/except =====
    def _exec_try(self, node: ast.Try):
        try:
            retval = self._exec(node.body)
            if isinstance(retval, (_Return, _Continue, _Break)):
                return retval
        except Exception as exc:
            if isinstance(exc, WrappedException):
                exc = exc.original
            # draconic diff: limit errors cannot be caught
            if isinstance(exc, LimitException):
                raise
            # enter into the except handlers
            for handler in node.handlers:
                if self._except_handler_matches(handler, exc):
                    retval = self._except_handler(handler)
                    if isinstance(retval, (_Return, _Continue, _Break)):
                        return retval
                    break
            else:
                raise
        else:
            retval = self._exec(node.orelse)
            if isinstance(retval, (_Return, _Continue, _Break)):
                return retval
        finally:
            retval = self._exec(node.finalbody)
            if isinstance(retval, (_Return, _Continue, _Break)):
                return retval

    def _except_handler_matches(self, node: ast.ExceptHandler, exc: BaseException) -> bool:
        # draconic diff: exception handlers must be string literals, tuple[str] literals, or bare
        if node.type is None:
            return True
        if isinstance(node.type, ast.Str):
            return type(exc).__name__ == self._eval_str(node.type)
        elif isinstance(node.type, ast.Tuple):
            if not all(isinstance(item, ast.Str) for item in node.type.elts):
                raise FeatureNotAvailable(
                    "'except' clause expressions must be string literals or tuple of string literals", node, self._expr
                )
            return type(exc).__name__ in self._eval_tuple(node.type)
        else:
            raise FeatureNotAvailable(
                "'except' clause expressions must be string literals or tuple of string literals", node, self._expr
            )

    def _except_handler(self, node: ast.ExceptHandler):
        # draconic diff: 'as X' clause not allowed
        if node.name is not None:
            raise FeatureNotAvailable("'except ... as X' is not available in this interpreter", node, self._expr)
        # run body
        return self._exec(node.body)



from __future__ import annotations

import ast
import inspect
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Optional, Tuple

from lsprotocol import types

from .runtime import _default_builtins


@dataclass
class FunctionSig:
    name: str
    params: List[str]
    doc: str = ""

    @property
    def label(self) -> str:
        params = ", ".join(self.params)
        return f"{self.name}({params})"


def load_signatures() -> Dict[str, FunctionSig]:
    sigs: dict[str, FunctionSig] = {}
    sigs.update(_builtin_sigs())
    sigs.update(_runtime_helper_sigs())
    sigs.update(_avrae_function_sigs())
    return sigs


def _builtin_sigs() -> Dict[str, FunctionSig]:
    sigs: dict[str, FunctionSig] = {}
    for name, fn in _default_builtins().items():
        try:
            sig = inspect.signature(fn)
        except (TypeError, ValueError):
            continue
        params = [p.name for p in sig.parameters.values()]
        sigs[name] = FunctionSig(name=name, params=params, doc=fn.__doc__ or "")
    return sigs


def _runtime_helper_sigs() -> Dict[str, FunctionSig]:
    helpers = {
        "get_gvar": ["address"],
        "get_svar": ["name", "default=None"],
        "get_cvar": ["name", "default=None"],
        "get_uvar": ["name", "default=None"],
        "get_uvars": [],
        "set_uvar": ["name", "value"],
        "set_uvar_nx": ["name", "value"],
        "delete_uvar": ["name"],
        "uvar_exists": ["name"],
        "exists": ["name"],
        "get": ["name", "default=None"],
        "using": ["**imports"],
        "signature": ["data=0"],
        "verify_signature": ["sig"],
        "print": ["*values"],
        "character": [],
        "combat": [],
        "argparse": ["args", "character=None", "splitter=argsplit", "parse_ephem=True"],
    }
    return {name: FunctionSig(name=name, params=params) for name, params in helpers.items()}


def _avrae_function_sigs() -> Dict[str, FunctionSig]:
    sigs: dict[str, FunctionSig] = {}
    module_path = Path(__file__).resolve().parent.parent / "avrae" / "aliasing" / "api" / "functions.py"
    if not module_path.exists():
        return sigs
    try:
        tree = ast.parse(module_path.read_text())
    except Exception:
        return sigs

    for node in tree.body:
        if isinstance(node, ast.FunctionDef) and not node.name.startswith("_"):
            params: list[str] = []
            defaults = list(node.args.defaults)
            default_offset = len(node.args.args) - len(defaults)
            for idx, arg in enumerate(node.args.args):
                default_val = None
                if idx >= default_offset:
                    default_node = defaults[idx - default_offset]
                    try:
                        default_val = ast.literal_eval(default_node)
                    except Exception:
                        default_val = None
                params.append(f"{arg.arg}={default_val}" if default_val is not None else arg.arg)
            doc = ast.get_docstring(node) or ""
            sigs[node.name] = FunctionSig(name=node.name, params=params, doc=doc)
    return sigs


def signature_help_for_code(code: str, line: int, character: int, sigs: Dict[str, FunctionSig]) -> Optional[types.SignatureHelp]:
    try:
        tree = ast.parse(code)
    except SyntaxError:
        return None

    target_call: ast.Call | None = None
    for node in ast.walk(tree):
        if isinstance(node, ast.Call):
            if hasattr(node, "lineno") and hasattr(node, "col_offset"):
                start = (node.lineno - 1, node.col_offset)
                end_line = getattr(node, "end_lineno", node.lineno) - 1
                end_col = getattr(node, "end_col_offset", node.col_offset)
                if _pos_within((line, character), start, (end_line, end_col)):
                    target_call = node
                    break

    if not target_call:
        return None

    if isinstance(target_call.func, ast.Name):
        name = target_call.func.id
    else:
        return None

    if name not in sigs:
        return None

    fsig = sigs[name]
    sig_info = types.SignatureInformation(
        label=fsig.label,
        documentation=fsig.doc,
        parameters=[types.ParameterInformation(label=p) for p in fsig.params],
    )
    active_param = min(len(target_call.args), max(len(fsig.params) - 1, 0))
    return types.SignatureHelp(signatures=[sig_info], active_signature=0, active_parameter=active_param)


def _pos_within(pos: Tuple[int, int], start: Tuple[int, int], end: Tuple[int, int]) -> bool:
    (line, col) = pos
    (sl, sc) = start
    (el, ec) = end
    if line < sl or line > el:
        return False
    if line == sl and col < sc:
        return False
    if line == el and col > ec:
        return False
    return True



from __future__ import annotations

import json
import logging
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, Iterable

import httpx

from .config import AvraeLSConfig, ContextProfile, VarSources
from .cvars import derive_character_cvars

log = logging.getLogger(__name__)


@dataclass
class ContextData:
    ctx: Dict[str, Any] = field(default_factory=dict)
    combat: Dict[str, Any] = field(default_factory=dict)
    character: Dict[str, Any] = field(default_factory=dict)
    vars: VarSources = field(default_factory=VarSources)


class ContextBuilder:
    def __init__(self, config: AvraeLSConfig):
        self._config = config
        self._gvar_resolver = GVarResolver(config)

    @property
    def gvar_resolver(self) -> "GVarResolver":
        return self._gvar_resolver

    def build(self, profile_name: str | None = None) -> ContextData:
        profile = self._select_profile(profile_name)
        merged_vars = self._merge_character_cvars(profile.character, self._load_var_files().merge(profile.vars))
        self._gvar_resolver.seed(merged_vars.gvars)
        return ContextData(
            ctx=dict(profile.ctx),
            combat=dict(profile.combat),
            character=dict(profile.character),
            vars=merged_vars,
        )

    def _select_profile(self, profile_name: str | None) -> ContextProfile:
        if profile_name and profile_name in self._config.profiles:
            return self._config.profiles[profile_name]
        if self._config.default_profile in self._config.profiles:
            return self._config.profiles[self._config.default_profile]
        return next(iter(self._config.profiles.values()))

    def _load_var_files(self) -> VarSources:
        merged = VarSources()
        for path in self._config.var_files:
            data = _read_json_file(path)
            if data is None:
                continue
            merged = merged.merge(VarSources.from_data(data))
        return merged

    def _merge_character_cvars(self, character: Dict[str, Any], vars: VarSources) -> VarSources:
        merged = vars
        char_cvars = character.get("cvars") or {}
        if char_cvars:
            merged = merged.merge(VarSources(cvars=dict(char_cvars)))

        builtin_cvars = derive_character_cvars(character)
        if builtin_cvars:
            merged = merged.merge(VarSources(cvars=builtin_cvars))
        return merged


class GVarResolver:
    def __init__(self, config: AvraeLSConfig):
        self._config = config
        self._cache: Dict[str, Any] = {}

    def reset(self, gvars: Dict[str, Any] | None = None) -> None:
        self._cache = {}
        if gvars:
            self._cache.update({str(k): v for k, v in gvars.items()})

    def seed(self, gvars: Dict[str, Any] | None = None) -> None:
        """Merge provided gvars into the cache without dropping fetched values."""
        if not gvars:
            return
        for k, v in gvars.items():
            self._cache[str(k)] = v

    def get_local(self, key: str) -> Any:
        return self._cache.get(str(key))

    async def ensure(self, key: str) -> bool:
        key = str(key)
        if key in self._cache:
            return True
        if not self._config.enable_gvar_fetch:
            return False
        if not self._config.service.token:
            log.debug("GVAR fetch skipped for %s: no token configured", key)
            return False

        base_url = self._config.service.base_url.rstrip("/")
        url = f"{base_url}/gvars/{key}"
        headers = {"Authorization": f"Bearer {self._config.service.token}"}
        try:
            async with httpx.AsyncClient(timeout=5) as client:
                resp = await client.get(url, headers=headers)
        except Exception as exc:
            log.debug("GVAR fetch failed for %s: %s", key, exc)
            return False

        if resp.status_code != 200:
            log.debug("GVAR fetch returned %s for %s", resp.status_code, key)
            return False

        payload = resp.json()
        value = payload.get("value") or payload.get("content") or payload.get("body")
        if value is None:
            log.debug("GVAR %s payload missing value", key)
            return False
        self._cache[key] = value
        return True

    def snapshot(self) -> Dict[str, Any]:
        return dict(self._cache)

    async def refresh(self, seed: Dict[str, Any] | None = None, keys: Iterable[str] | None = None) -> Dict[str, Any]:
        self.reset(seed)
        if keys:
            for key in keys:
                await self.ensure(key)
        return self.snapshot()


def _read_json_file(path: Path) -> Dict[str, Any] | None:
    try:
        text = path.read_text()
    except FileNotFoundError:
        log.debug("Var file not found: %s", path)
        return None
    except OSError as exc:
        log.warning("Failed to read var file %s: %s", path, exc)
        return None

    try:
        return json.loads(text)
    except json.JSONDecodeError as exc:
        log.warning("Failed to parse var file %s: %s", path, exc)
        return None



import abc
import ast

from .versions import PY_310

__all__ = (
    "DraconicException",
    "DraconicSyntaxError",
    "InvalidExpression",
    "NotDefined",
    "FeatureNotAvailable",
    "DraconicValueError",
    "LimitException",
    "NumberTooHigh",
    "IterableTooLong",
    "TooManyStatements",
    "TooMuchRecursion",
    "WrappedException",
    "AnnotatedException",
    "NestedException",
    "_PostponedRaise",
    "_raise_in_context",
)


class DraconicException(Exception):
    """Base exception for all exceptions in this library."""

    __drac_context__: str = None

    def __init__(self, msg):
        super().__init__(msg)
        self.msg = msg


class DraconicSyntaxError(DraconicException):
    """Bad syntax."""

    def __init__(self, original: SyntaxError, expr):
        super().__init__(original.msg)
        self.lineno = original.lineno
        self.offset = original.offset
        self.end_lineno = None
        self.end_offset = None
        self.expr = expr

        if PY_310:
            self.end_lineno = original.end_lineno
            self.end_offset = original.end_offset

    @classmethod
    def from_node(cls, node: ast.AST, msg: str, expr):
        if PY_310:
            inner = SyntaxError(
                msg, ("<string>", node.lineno, node.col_offset + 1, expr, node.end_lineno, node.end_col_offset + 1)
            )
        else:
            inner = SyntaxError(msg, ("<string>", node.lineno, node.col_offset + 1, expr))
        return cls(inner, expr)


class InvalidExpression(DraconicException):
    """Base exception for all exceptions during run-time."""

    def __init__(self, msg, node, expr):
        super().__init__(msg)
        self.node = node
        self.expr = expr


class NotDefined(InvalidExpression):
    """Some name is not defined."""

    pass


class FeatureNotAvailable(InvalidExpression):
    """What you're trying to do is not allowed."""

    pass


class DraconicValueError(InvalidExpression):
    """Bad value passed to some function."""

    pass


class LimitException(InvalidExpression):
    """
    Something exceeded execution limits.
    """

    pass


class NumberTooHigh(LimitException):
    """Some number is way too big."""

    pass


class IterableTooLong(LimitException):
    """Some iterable is way too big."""

    pass


class TooManyStatements(LimitException):
    """Tried to execute too many statements."""

    pass


class TooMuchRecursion(LimitException):
    """Too deep in recursion."""

    pass


class WrappedException(InvalidExpression, abc.ABC):
    """abstract base exception for a lib exception that wraps some other exception"""

    original: BaseException


class AnnotatedException(WrappedException):
    """A wrapper for another exception to handle lineno info."""

    def __init__(self, original, node, expr):
        super().__init__(str(original), node, expr)
        self.original = original


class NestedException(WrappedException):
    """An exception occurred in a user-defined function call."""

    def __init__(self, msg, node, expr, last_exc):
        super().__init__(msg, node, expr)
        self.last_exc = last_exc  # type: DraconicException  # used for tracebacking
        # keep a reference to the end of the chain for easy comparison
        if isinstance(last_exc, WrappedException):
            self.original = last_exc.original
        else:
            self.original = last_exc


# we need to raise some exception, but don't have the node context right now
class _PostponedRaise(Exception):
    def __init__(self, cls, *args, **kwargs):
        self.cls = cls
        self.args = args
        self.kwargs = kwargs


def _raise_in_context(cls, *args, **kwargs):
    raise _PostponedRaise(cls, *args, **kwargs)



from __future__ import annotations

import re
from dataclasses import dataclass
from typing import List


@dataclass
class DraconicBlock:
    code: str
    line_offset: int
    char_offset: int = 0
    line_count: int = 0


DRACONIC_RE = re.compile(r"<drac2>([\s\S]*?)</drac2>", re.IGNORECASE)


def find_draconic_blocks(source: str) -> List[DraconicBlock]:
    blocks: list[DraconicBlock] = []
    for match in DRACONIC_RE.finditer(source):
        raw = match.group(1)
        prefix = source[: match.start()]
        line_offset = prefix.count("\n")
        line_count = raw.count("\n") + 1 if raw else 1
        # Trim leading blank lines inside the block while tracking the line shift
        while raw.startswith("\n"):
            raw = raw[1:]
            line_offset += 1
        blocks.append(DraconicBlock(code=raw, line_offset=line_offset, char_offset=0, line_count=line_count))
    return blocks


def primary_block_or_source(source: str) -> tuple[str, int, int]:
    blocks = find_draconic_blocks(source)
    if not blocks:
        return source, 0, 0
    block = blocks[0]
    return block.code, block.line_offset, block.char_offset



from __future__ import annotations

import re
from typing import List, Sequence


_NUM_PLACEHOLDER_RE = re.compile(r"%(\d+)%|&(\d+)&")


def _escape_quotes(arg: str) -> str:
    """Escape quotes/backslashes to mimic Avrae's quote-escaping behavior."""
    return arg.replace("\\", "\\\\").replace('"', '\\"').replace("'", "\\'")


def _ensure_args(args: Sequence[str] | None, count: int) -> List[str]:
    out = list(args or [])
    while len(out) < count:
        out.append(f"arg{len(out) + 1}")
    return out


def apply_argument_parsing(text: str, args: Sequence[str] | None = None) -> str:
    """
    Apply Avrae's argument placeholder replacement rules to an alias body.

    Supports:
    - %N%   : non-code replacement (quotes added if arg contains spaces)
    - %*%   : full arg string
    - &N&   : in-code replacement with quote escaping
    - &*&   : full arg string with quote escaping
    - &ARGS&: Python-style list literal of args
    """
    max_idx = 0
    for match in _NUM_PLACEHOLDER_RE.finditer(text):
        groups = [g for g in match.groups() if g]
        if groups:
            max_idx = max(max_idx, int(groups[0]))

    args_list = _ensure_args(args, max_idx)

    def get_arg(idx: int) -> str:
        zero = idx - 1
        if zero < 0 or zero >= len(args_list):
            return ""
        return str(args_list[zero])

    def percent_repl(match: re.Match) -> str:
        idx = int(match.group(1))
        val = get_arg(idx)
        return f'"{val}"' if " " in val else val

    def amp_repl(match: re.Match) -> str:
        idx = int(match.group(1))
        val = get_arg(idx)
        return _escape_quotes(val)

    full_args = " ".join(args_list)
    escaped_full_args = _escape_quotes(full_args)
    list_literal = "[" + ", ".join(repr(a) for a in args_list) + "]"

    # order: numeric placeholders first, then multi-arg macros
    text = re.sub(r"%(\d+)%", percent_repl, text)
    text = text.replace("%*%", full_args)
    text = re.sub(r"&(\d+)&", amp_repl, text)
    text = text.replace("&*&", escaped_full_args)
    text = text.replace("&ARGS&", list_literal)
    return text



from __future__ import annotations

import json
import logging
import math
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, Iterable, Tuple

CONFIG_FILENAME = ".avraels.json"
log = logging.getLogger(__name__)


class ConfigError(Exception):
    """Raised when a workspace config file cannot be parsed."""


@dataclass
class DiagnosticSettings:
    semantic_level: str = "warning"
    runtime_level: str = "error"


@dataclass
class AvraeServiceConfig:
    base_url: str = "https://api.avrae.io"
    token: str | None = None


@dataclass
class VarSources:
    cvars: Dict[str, Any] = field(default_factory=dict)
    uvars: Dict[str, Any] = field(default_factory=dict)
    svars: Dict[str, Any] = field(default_factory=dict)
    gvars: Dict[str, Any] = field(default_factory=dict)

    @classmethod
    def from_data(cls, data: Dict[str, Any] | None) -> "VarSources":
        data = data or {}
        return cls(
            cvars=dict(data.get("cvars") or {}),
            uvars=dict(data.get("uvars") or {}),
            svars=dict(data.get("svars") or {}),
            gvars=dict(data.get("gvars") or {}),
        )

    def merge(self, other: "VarSources") -> "VarSources":
        def _merge(lhs: Dict[str, Any], rhs: Dict[str, Any]) -> Dict[str, Any]:
            merged = dict(lhs)
            merged.update(rhs)
            return merged

        return VarSources(
            cvars=_merge(self.cvars, other.cvars),
            uvars=_merge(self.uvars, other.uvars),
            svars=_merge(self.svars, other.svars),
            gvars=_merge(self.gvars, other.gvars),
        )

    def to_initial_names(self) -> Dict[str, Any]:
        names: Dict[str, Any] = {}
        # Bind uvars first so cvars take precedence, matching Avrae's local > cvar > uvar lookup.
        names.update(self.uvars)
        names.update(self.cvars)
        return names


@dataclass
class ContextProfile:
    name: str
    ctx: Dict[str, Any] = field(default_factory=dict)
    combat: Dict[str, Any] = field(default_factory=dict)
    character: Dict[str, Any] = field(default_factory=dict)
    vars: VarSources = field(default_factory=VarSources)
    description: str = ""


@dataclass
class AvraeLSConfig:
    workspace_root: Path
    enable_gvar_fetch: bool = False
    service: AvraeServiceConfig = field(default_factory=AvraeServiceConfig)
    var_files: Tuple[Path, ...] = field(default_factory=tuple)
    default_profile: str = "default"
    profiles: Dict[str, ContextProfile] = field(default_factory=dict)
    diagnostics: DiagnosticSettings = field(default_factory=DiagnosticSettings)

    @classmethod
    def default(cls, workspace_root: Path) -> "AvraeLSConfig":
        abilities = {
            "strength": 16,
            "dexterity": 14,
            "constitution": 15,
            "intelligence": 10,
            "wisdom": 12,
            "charisma": 13,
            "prof_bonus": 3,
        }

        def _mod(score: int) -> int:
            return math.floor((score - 10) / 2)

        skill_profs = {"athletics", "perception", "stealth", "survival"}
        skills = {}
        for name, ability in {
            "acrobatics": "dexterity",
            "animalHandling": "wisdom",
            "arcana": "intelligence",
            "athletics": "strength",
            "deception": "charisma",
            "history": "intelligence",
            "initiative": "dexterity",
            "insight": "wisdom",
            "intimidation": "charisma",
            "investigation": "intelligence",
            "medicine": "wisdom",
            "nature": "intelligence",
            "perception": "wisdom",
            "performance": "charisma",
            "persuasion": "charisma",
            "religion": "intelligence",
            "sleightOfHand": "dexterity",
            "stealth": "dexterity",
            "survival": "wisdom",
            "strength": "strength",
            "dexterity": "dexterity",
            "constitution": "constitution",
            "intelligence": "intelligence",
            "wisdom": "wisdom",
            "charisma": "charisma",
        }.items():
            base = _mod(abilities[ability])
            prof = 1 if name in skill_profs else 0
            skills[name] = {"value": base + abilities["prof_bonus"] * prof, "prof": prof, "bonus": 0, "adv": None}

        saves = {
            "str": _mod(abilities["strength"]) + abilities["prof_bonus"],
            "dex": _mod(abilities["dexterity"]),
            "con": _mod(abilities["constitution"]) + abilities["prof_bonus"],
            "int": _mod(abilities["intelligence"]),
            "wis": _mod(abilities["wisdom"]),
            "cha": _mod(abilities["charisma"]),
        }

        attacks = [
            {
                "name": "Longsword",
                "verb": "swings",
                "proper": False,
                "activation_type": 1,
                "raw": {"name": "Longsword", "bonus": "+7", "damage": "1d8+4 slashing"},
            },
            {
                "name": "Shortbow",
                "verb": "looses",
                "proper": False,
                "activation_type": 1,
                "raw": {"name": "Shortbow", "bonus": "+6", "damage": "1d6+3 piercing"},
            },
        ]

        spellbook = {
            "dc": 14,
            "sab": 6,
            "caster_level": 5,
            "spell_mod": 3,
            "pact_slot_level": None,
            "num_pact_slots": None,
            "max_pact_slots": None,
            "slots": {1: 4, 2: 2},
            "max_slots": {1: 4, 2: 2},
            "spells": [
                {"name": "Cure Wounds", "dc": None, "sab": None, "mod": None, "prepared": True},
                {"name": "Hunter's Mark", "dc": None, "sab": None, "mod": None, "prepared": True},
                {"name": "Fire Bolt", "dc": None, "sab": None, "mod": None, "prepared": True},
            ],
        }

        consumables = {
            "Hit Dice": {
                "name": "Hit Dice",
                "value": 5,
                "max": 5,
                "min": 0,
                "reset_on": "long",
                "display_type": None,
                "reset_to": None,
                "reset_by": None,
                "title": "d10 hit dice",
                "desc": "Hit dice pool",
            },
            "Bardic Inspiration": {
                "name": "Bardic Inspiration",
                "value": 3,
                "max": 3,
                "min": 0,
                "reset_on": "long",
                "display_type": "bubble",
                "reset_to": "max",
                "reset_by": None,
                "title": None,
                "desc": None,
            },
        }

        character = {
            "name": "Aelar Wyn",
            "race": "Half-Elf",
            "background": "Outlander",
            "description": "Scout of the Emerald Enclave.",
            "image": "https://example.invalid/aelar.png",
            "owner": 1010101010,
            "upstream": "char_aelar_wyn",
            "sheet_type": "beyond",
            "creature_type": "humanoid",
            "stats": abilities,
            "levels": {"Fighter": 3, "Rogue": 2},
            "attacks": attacks,
            "actions": [
                {
                    "name": "Second Wind",
                    "activation_type": 3,
                    "activation_type_name": "BONUS_ACTION",
                    "description": "+1d10+5 hp",
                    "snippet": "+1d10+5 hp",
                },
                {
                    "name": "Action Surge",
                    "activation_type": 1,
                    "activation_type_name": "ACTION",
                    "description": "Take one additional action.",
                    "snippet": "Take one additional action.",
                },
            ],
            "skills": skills,
            "saves": saves,
            "resistances": {"resist": [{"dtype": "fire", "unless": [], "only": []}], "vuln": [], "immune": [], "neutral": []},
            "spellbook": spellbook,
            "consumables": consumables,
            "cvars": {"favorite_enemy": "goblinoids", "fighting_style": "defense"},
            "coinpurse": {"pp": 1, "gp": 47, "ep": 0, "sp": 12, "cp": 34},
            "death_saves": {"successes": 0, "fails": 0},
            "max_hp": 44,
            "hp": 38,
            "temp_hp": 3,
            "ac": 17,
            "passive_perception": 15,
            "speed": 30,
            "class_levels": {"Fighter": 3, "Rogue": 2},
            "csettings": {"compact_coins": False},
        }

        me_combatant = {
            "name": character["name"],
            "id": "cmb_aelar",
            "controller": character["owner"],
            "group": None,
            "race": character["race"],
            "monster_name": None,
            "is_hidden": False,
            "init": 18,
            "initmod": 4,
            "type": "combatant",
            "note": "On watch",
            "effects": [
                {
                    "name": "Hunter's Mark",
                    "duration": 600,
                    "remaining": 540,
                    "desc": "Mark one target; deal +1d6 damage to it.",
                    "concentration": True,
                    "combatant_name": character["name"],
                }
            ],
            "stats": character["stats"],
            "levels": character["levels"],
            "skills": character["skills"],
            "saves": character["saves"],
            "resistances": character["resistances"],
            "spellbook": character["spellbook"],
            "attacks": character["attacks"],
            "max_hp": character["max_hp"],
            "hp": character["hp"],
            "temp_hp": character["temp_hp"],
            "ac": character["ac"],
            "creature_type": character["creature_type"],
        }

        goblin = {
            "name": "Goblin Cutter",
            "id": "cmb_gob1",
            "controller": None,
            "group": "Goblins",
            "race": None,
            "monster_name": "Goblin",
            "is_hidden": False,
            "init": 12,
            "initmod": 2,
            "type": "combatant",
            "note": "",
            "effects": [],
            "stats": {"strength": 8, "dexterity": 14, "constitution": 10, "intelligence": 8, "wisdom": 10, "charisma": 8, "prof_bonus": 2},
            "levels": {},
            "skills": {"stealth": {"value": 6, "prof": 1, "bonus": 0, "adv": None}},
            "saves": {"dex": 4},
            "resistances": {"resist": [], "vuln": [], "immune": [], "neutral": []},
            "spellbook": {"spells": []},
            "attacks": [{"name": "Scimitar", "verb": "slashes", "proper": False, "activation_type": 1, "raw": {"name": "Scimitar", "bonus": "+4", "damage": "1d6+2 slashing"}}],
            "max_hp": 11,
            "hp": 11,
            "temp_hp": 0,
            "ac": 15,
            "creature_type": "humanoid",
        }

        group = {"name": "Goblins", "id": "grp_goblins", "init": 12, "type": "group", "combatants": [goblin]}

        combat = {
            "name": "Goblin Ambush",
            "round_num": 2,
            "turn_num": 15,
            "combatants": [me_combatant, goblin],
            "groups": [group],
            "me": me_combatant,
            "current": group,
            "metadata": {"scene": "Forest road", "weather": "light rain"},
        }

        ctx = {
            "guild": {"id": 123456789012345678, "name": "Fabled Realms"},
            "channel": {
                "id": 234567890123456789,
                "name": "tavern-rp",
                "topic": "Adventuring party chat",
                "category": {"id": 9876543210, "name": "Adventures"},
                "parent": None,
            },
            "author": {
                "id": 345678901234567890,
                "name": "AelarW",
                "discriminator": "0420",
                "display_name": "Aelar Wyn",
                "roles": [{"id": 4567, "name": "DM"}, {"id": 4568, "name": "Player"}],
            },
            "prefix": "!",
            "alias": "mockalias",
            "message_id": 456789012345678901,
        }

        profile = ContextProfile(
            name="default",
            ctx=ctx,
            combat=combat,
            character=character,
            description="Built-in Avrae LS mock profile with realistic sample data.",
        )
        return cls(
            workspace_root=workspace_root,
            profiles={"default": profile},
        )


def load_config(workspace_root: Path) -> Tuple[AvraeLSConfig, Iterable[str]]:
    """Load `.avraels.json` from the workspace root, returning config and warnings."""
    path = workspace_root / CONFIG_FILENAME
    if not path.exists():
        return AvraeLSConfig.default(workspace_root), []

    try:
        raw = json.loads(path.read_text())
    except json.JSONDecodeError as exc:
        warning = f"Failed to parse {CONFIG_FILENAME}: {exc}"
        log.warning(warning)
        return AvraeLSConfig.default(workspace_root), [warning]

    warnings: list[str] = []
    enable_gvar_fetch = bool(raw.get("enableGvarFetch", False))

    service_cfg = raw.get("avraeService") or {}
    service = AvraeServiceConfig(
        base_url=str(service_cfg.get("baseUrl") or AvraeServiceConfig.base_url),
        token=service_cfg.get("token"),
    )

    diag_cfg = raw.get("diagnostics") or {}
    diagnostics = DiagnosticSettings(
        semantic_level=str(diag_cfg.get("semanticLevel") or "warning").lower(),
        runtime_level=str(diag_cfg.get("runtimeLevel") or "error").lower(),
    )

    var_files = tuple(
        _resolve_var_file(workspace_root, file_path)
        for file_path in raw.get("varFiles", [])
        if isinstance(file_path, str)
    )

    profiles: Dict[str, ContextProfile] = {}
    raw_profiles = raw.get("profiles") or {}
    for name, data in raw_profiles.items():
        profiles[name] = ContextProfile(
            name=name,
            ctx=dict(data.get("ctx") or {}),
            combat=dict(data.get("combat") or {}),
            character=dict(data.get("character") or {}),
            vars=VarSources.from_data(data.get("vars")),
            description=str(data.get("description") or ""),
        )

    default_profile = str(raw.get("defaultProfile") or "default")
    if default_profile not in profiles and profiles:
        warnings.append(
            f"defaultProfile '{default_profile}' not found; falling back to first profile in file."
        )
        default_profile = next(iter(profiles))

    if not profiles:
        base = AvraeLSConfig.default(workspace_root)
        profiles = base.profiles
        default_profile = base.default_profile

    cfg = AvraeLSConfig(
        workspace_root=workspace_root,
        enable_gvar_fetch=enable_gvar_fetch,
        service=service,
        var_files=var_files,
        default_profile=default_profile,
        profiles=profiles,
        diagnostics=diagnostics,
    )
    return cfg, warnings


def _resolve_var_file(root: Path, file_path: str) -> Path:
    candidate = Path(file_path)
    if not candidate.is_absolute():
        candidate = root / candidate
    return candidate



from __future__ import annotations

import re
import shlex
from dataclasses import dataclass
from typing import Any, Optional, Tuple

from .parser import DRACONIC_RE
from .runtime import ExecutionResult, MockExecutor
from .context import ContextData, GVarResolver
from .argument_parsing import apply_argument_parsing


@dataclass
class RenderedAlias:
    command: str
    stdout: str
    error: Optional[BaseException]
    last_value: Any | None = None


def _strip_alias_header(text: str) -> str:
    lines = text.splitlines()
    if lines and lines[0].lstrip().startswith("!alias"):
        first = lines[0].lstrip()
        parts = first.split(maxsplit=2)
        remainder = parts[2] if len(parts) > 2 else ""
        body = "\n".join(lines[1:])
        if remainder:
            return remainder + ("\n" + body if body else "")
        return body
    return text


async def render_alias_command(
    text: str,
    executor: MockExecutor,
    ctx_data: ContextData,
    resolver: GVarResolver,
    args: list[str] | None = None,
) -> RenderedAlias:
    """Replace <drac2> blocks with their evaluated values and return final command."""
    body = _strip_alias_header(text)
    body = apply_argument_parsing(body, args)
    stdout_parts: list[str] = []
    parts: list[str] = []
    last_value = None
    error: BaseException | None = None

    pos = 0
    for match in DRACONIC_RE.finditer(body):
        parts.append(body[pos: match.start()])
        code = match.group(1)
        result: ExecutionResult = await executor.run(code, ctx_data, resolver)
        if result.stdout:
            stdout_parts.append(result.stdout)
        if result.error:
            error = result.error
            break
        last_value = result.value
        parts.append("" if result.value is None else str(result.value))
        pos = match.end()

    if error is None:
        parts.append(body[pos:])

    final_command = "".join(parts)
    return RenderedAlias(command=final_command, stdout="".join(stdout_parts), error=error, last_value=last_value)


def validate_embed_payload(payload: str) -> Tuple[bool, str | None]:
    """
    Light validation for embed previews using Avrae-style flags.

    Accepts strings such as "-title Foo -f \"T|Body\"" and validates arguments.
    Returns (is_valid, error_message) without attempting to parse JSON objects.
    """
    text = payload.strip()
    if not text:
        return False, "Embed payload is empty."

    return _validate_embed_flags(text)


def _validate_embed_flags(text: str) -> Tuple[bool, str | None]:
    """Validate embed flags according to Avrae's help text."""
    if not text:
        return False, "Embed payload is empty."

    try:
        tokens = shlex.split(text)
    except ValueError as exc:  # pragma: no cover - defensive only
        return False, f"Embed payload could not be parsed: {exc}"

    flag_handlers = {
        "-title": lambda val: _require_value("-title", val),
        "-desc": lambda val: _require_value("-desc", val),
        "-thumb": lambda val: _require_value("-thumb", val),
        "-image": lambda val: _require_value("-image", val),
        "-footer": lambda val: _require_value("-footer", val),
        "-f": _validate_field_arg,
        "-color": _validate_color_arg,
        "-t": _validate_timeout_arg,
    }

    i = 0
    while i < len(tokens):
        tok = tokens[i]
        key = tok.lower()
        if not tok.startswith("-"):
            i += 1
            continue
        if key not in flag_handlers:
            return False, f"Embed payload contains unknown flag '{tok}'."
        next_val = tokens[i + 1] if i + 1 < len(tokens) else None
        ok, err, consumed = flag_handlers[key](next_val)
        if not ok:
            return False, err
        i += consumed + 1
    return True, None


def _require_value(flag: str, value: str | None) -> Tuple[bool, str | None, int]:
    if value is None or value.startswith("-"):
        return False, f"Embed flag '{flag}' requires a value.", 0
    return True, None, 1


def _validate_field_arg(value: str | None) -> Tuple[bool, str | None, int]:
    ok, err, consumed = _require_value("-f", value)
    if not ok:
        return ok, err, consumed
    assert value is not None  # for type checker
    parts = value.split("|")
    if len(parts) < 2 or len(parts) > 3:
        return False, "Embed field must be in the form \"Title|Text[|inline]\".", consumed
    if not parts[0] or not parts[1]:
        return False, "Embed field title and text cannot be empty.", consumed
    if len(parts) == 3 and parts[2].lower() not in ("inline", ""):
        return False, "Embed field inline value must be 'inline' or omitted.", consumed
    return True, None, consumed


def _validate_color_arg(value: str | None) -> Tuple[bool, str | None, int]:
    if value is None or value.startswith("-"):
        # Random color is allowed when omitted
        return True, None, 0
    if not re.match(r"^(?:#|0x)?[0-9a-fA-F]{6}$", value):
        return False, "Embed color must be a 6-hex value (e.g. #ff00ff).", 1
    return True, None, 1


def _validate_timeout_arg(value: str | None) -> Tuple[bool, str | None, int]:
    ok, err, consumed = _require_value("-t", value)
    if not ok:
        return ok, err, consumed
    assert value is not None  # for type checker
    try:
        num = int(value)
    except ValueError:
        return False, "Embed timeout (-t) must be an integer.", consumed
    if num < 0 or num > 600:
        return False, "Embed timeout (-t) must be between 0 and 600 seconds.", consumed
    return True, None, consumed


def simulate_command(command: str) -> tuple[str | None, str | None, str | None]:
    """Very small shim to preview common commands."""
    text = command.strip()
    if not text:
        return None, None, None
    head, *rest = text.split(maxsplit=1)
    payload = rest[0] if rest else ""
    lowered = head.lower()
    if lowered == "echo":
        return payload, "echo", None
    if lowered == "embed":
        valid, error = validate_embed_payload(payload)
        return payload, "embed", error
    return None, head, None



"""
This submodule contains helpful public utilities.
"""

import textwrap
from collections import namedtuple
from typing import List, Union

from .exceptions import AnnotatedException, DraconicException, DraconicSyntaxError, InvalidExpression, NestedException

LineInfo = namedtuple("LineInfo", "lineno col_offset end_lineno end_col_offset")


def format_traceback(exc: DraconicException) -> List[str]:
    """
    Given an exception raised by this library during execution of a userscript, provide a ``traceback``-like format
    of the call stack leading to the exception.
    """
    tb = ["Traceback (most recent call last):\n"]

    # show the call stack with pointers
    while isinstance(exc, NestedException):
        if exc.__drac_context__ is not None:
            tb.append(f"  Line {exc.node.lineno}, col {exc.node.col_offset}, in {exc.__drac_context__}\n")
        else:
            tb.append(f"  Line {exc.node.lineno}, col {exc.node.col_offset}\n")
        tb.append(textwrap.indent(format_exc_line_pointer(exc.node, exc.expr), "    "))

        exc = exc.last_exc

    # show the pointer to the original (draconic) exception
    in_func = f", in {exc.__drac_context__}" if exc.__drac_context__ is not None else ""
    if isinstance(exc, InvalidExpression):
        tb.append(f"  Line {exc.node.lineno}, col {exc.node.col_offset}{in_func}\n")
        tb.append(textwrap.indent(format_exc_line_pointer(extract_line_info(exc), exc.expr), "    "))
    elif isinstance(exc, DraconicSyntaxError):
        tb.append(f"  Line {exc.lineno}, col {exc.offset}{in_func}\n")
        tb.append(textwrap.indent(format_exc_line_pointer(extract_line_info(exc), exc.expr), "    "))
    else:  # pragma: no cover  # generic fallback, should never be hit
        tb.append(f"  While parsing expression{in_func}\n")

    # show the original exception message
    if isinstance(exc, AnnotatedException):
        exc = exc.original
    tb.append(f"{type(exc).__name__}: {exc!s}\n")
    return tb


def extract_line_info(exc: Union[InvalidExpression, DraconicSyntaxError]):
    if isinstance(exc, InvalidExpression):
        return LineInfo(
            exc.node.lineno,
            exc.node.col_offset,
            getattr(exc.node, "end_lineno", None),
            getattr(exc.node, "end_col_offset", None),
        )
    return LineInfo(
        exc.lineno,
        exc.offset - 1,
        exc.end_lineno,
        (exc.end_offset - 1) if exc.end_offset is not None else None,
    )


def format_exc_line_pointer(line_info: LineInfo, expr: str) -> str:
    the_line = expr.split("\n")[line_info.lineno - 1]

    # if the error spans multiple lines just point to the start
    if line_info.end_lineno is not None and line_info.end_lineno - line_info.lineno:
        return f"{the_line}\n{' ' * line_info.col_offset}^\n"

    # otherwise, if we have end_col_offset info, we need more than 1 carat
    if line_info.end_col_offset is not None:
        carats = "^" * (line_info.end_col_offset - line_info.col_offset)
    else:
        carats = "^"

    return textwrap.dedent(f"{the_line}\n{' ' * line_info.col_offset}{carats}\n")



from __future__ import annotations

import math
from typing import Any, Dict, Mapping

ABILITY_KEYS = ["strength", "dexterity", "constitution", "intelligence", "wisdom", "charisma"]
SAVE_KEYS = {
    "strength": "str",
    "dexterity": "dex",
    "constitution": "con",
    "intelligence": "int",
    "wisdom": "wis",
    "charisma": "cha",
}


def derive_character_cvars(character: Mapping[str, Any]) -> Dict[str, Any]:
    """Build the documented cvar table values from a character payload."""
    stats = character.get("stats") or {}
    saves = character.get("saves") or {}
    levels = character.get("levels") or {}
    spellbook = character.get("spellbook") or {}
    csettings = character.get("csettings") or {}

    cvars: dict[str, Any] = {}

    for ability in ABILITY_KEYS:
        score = _int_or_none(stats.get(ability))
        save_val = _int_or_none(saves.get(SAVE_KEYS[ability]))

        if score is not None:
            cvars[ability] = score
            cvars[f"{ability}Mod"] = math.floor((score - 10) / 2)
        if save_val is not None:
            cvars[f"{ability}Save"] = save_val

    armor = _int_or_none(character.get("ac"))
    if armor is not None:
        cvars["armor"] = armor

    description = character.get("description")
    if description is not None:
        cvars["description"] = description

    image = character.get("image")
    if image is not None:
        cvars["image"] = image

    name = character.get("name")
    if name is not None:
        cvars["name"] = name

    max_hp = _int_or_none(character.get("max_hp"))
    if max_hp is not None:
        cvars["hp"] = max_hp

    color = _color_hex(csettings.get("color"))
    if color is not None:
        cvars["color"] = color

    prof = _int_or_none(stats.get("prof_bonus"))
    if prof is not None:
        cvars["proficiencyBonus"] = prof

    spell_mod = _spell_mod(spellbook, prof)
    if spell_mod is not None:
        cvars["spell"] = spell_mod

    total_level = _sum_ints(levels.values())
    if total_level is not None:
        cvars["level"] = total_level

    for cls, lvl in levels.items():
        lvl_int = _int_or_none(lvl)
        if lvl_int is None:
            continue
        cvars[f"{str(cls).replace(' ', '')}Level"] = lvl_int

    return cvars


def _int_or_none(value: Any) -> int | None:
    if isinstance(value, dict) and "value" in value:
        value = value.get("value")
    try:
        return int(value)
    except (TypeError, ValueError):
        return None


def _spell_mod(spellbook: Mapping[str, Any], prof_bonus: int | None) -> int | None:
    if "spell_mod" in spellbook:
        return _int_or_none(spellbook.get("spell_mod"))
    if "sab" in spellbook and prof_bonus is not None:
        sab = _int_or_none(spellbook.get("sab"))
        if sab is not None:
            return sab - prof_bonus
    return None


def _sum_ints(values: Any) -> int | None:
    try:
        total = sum(int(v) for v in values)
    except Exception:
        return None
    return total


def _color_hex(value: Any) -> str | None:
    if value is None:
        return None
    try:
        return hex(int(value))[2:]
    except (TypeError, ValueError):
        return None
